from typing import List, Union

# Contraction path finding is offloaded to Cotengra
import cotengra as ctg
import numpy as np

# Underlying tensor objects can either be NumPy arrays or Sparse arrays
import sparse
from numpy.linalg import svd

# Qiskit quantum circuit integration
from qiskit import QuantumCircuit
from scipy.sparse.linalg import svds

from .tensor import Tensor

# Visualisation
from .visualisation import draw_arbitrary_tn, draw_quantum_circuit


class TensorNetwork:
    def __init__(
        self, tensors: List[Tensor], name: str = "TN", count_from: int = 1
    ) -> None:
        """
        Constructor for the TensorNetwork class.

        Args:
            tensors: A list of Tensor objects.
            name (optional): A name for the tensor network.

        Returns:
            A tensor network.
        """
        self.name = name
        i = count_from
        for t in tensors:
            t.labels.append(f"TN_T{i}")
            i += 1
        self.tensors = tensors
        self.indices = self.get_all_indices()

    def __str__(self) -> str:
        """
        Defines output of print.
        """
        output = "Tensor Network containing: \n"
        for t in self.tensors:
            shape = str(t.dimensions)
            indices = str(t.indices)
            output += f"Tensor with shape {shape} and indices {indices} \n"
        return output

    def __add__(self, other: "TensorNetwork") -> "TensorNetwork":
        """
        Defines addition for tensor networks.
        """
        current_tensor_number = len(self.tensors)
        self.tensors = self.tensors + other.tensors
        i = current_tensor_number
        for t in self.tensors[current_tensor_number:]:
            t.labels.append(f"TN_T{i}")
            i += 1
        self.indices = self.get_all_indices()
        return self

    @classmethod
    def from_qiskit_circuit(
        cls, qc: QuantumCircuit, dagger: bool = False
    ) -> "TensorNetwork":
        """
        Construct a tensor network from a Qiskit QuantumCircuit object.

        Args:
            qc: The QuantumCircuit.
            dagger: If True will construct a TN for the inverse circuit

        Returns:
            A tensor network.
        """
        num_qubits = qc.num_qubits
        index_prefixes = [f"QW{x}" for x in range(1, num_qubits + 1)]
        wire_counts = {str(x): 0 for x in range(num_qubits)}
        tensors = []

        data = qc.data
        if dagger:
            data = data[::-1]

        tensor_number = 1
        for inst in data:
            inst_num_qubits = inst.operation.num_qubits
            qidxs = [inst.qubits[i]._index for i in range(inst_num_qubits)]
            indices = [0] * (2 * len(qidxs))
            labels = [f"T{tensor_number}"]
            for qidx in qidxs:
                qw = index_prefixes[qidx]
                wire_in = str(wire_counts[str(qidx)])
                wire_out = str(wire_counts[str(qidx)] + 1)
                indices[qidxs.index(qidx)] = qw + "N" + wire_out
                indices[len(qidxs) + qidxs.index(qidx)] = qw + "N" + wire_in
                labels.append(f"Q{qidx}")
                wire_counts[str(qidx)] += 1
            inst_tensor = Tensor.from_qiskit_gate(inst, indices, labels, dagger)
            tensors.append(inst_tensor)
            tensor_number += 1

        tn = TensorNetwork(tensors, name="QuantumCircuit")
        return tn

    def get_index_to_tensor_dict(self) -> dict:
        """
        Build a dictionary mapping indices to their tensors.

        Returns:
            A dictionary of the form {idx : [tensor1,...]}
        """
        tn_dict = {}
        for t in self.tensors:
            for idx in t.indices:
                if idx in tn_dict:
                    tn_dict[idx].append(t)
                else:
                    tn_dict[idx] = [t]
        return tn_dict

    def get_label_to_tensor_dict(self) -> dict:
        """
        Build a dictionary mapping labels to their tensors.

        Returns:
            A dictionary of the form {label : [tensor1,...]}
        """
        tn_dict = {}
        for t in self.tensors:
            for label in t.labels:
                if label in tn_dict:
                    tn_dict[label].append(t)
                else:
                    tn_dict[label] = [t]
        return tn_dict

    def get_dimension_of_index(self, idx: str) -> int:
        """
        Get the dimension of an index.

        Args:
            idx: The index name.

        Returns:
            The dimension of idx.
        """
        for t in self.tensors:
            if idx in t.indices:
                return t.dimensions[t.indices.index(idx)]
        raise ValueError

    def get_internal_indices(self) -> List[str]:
        """
        Get the internal indices of the tensor network.

        Returns:
            Indices that are connected to 2 tensors in the network.
        """
        tn_dict = self.get_index_to_tensor_dict()
        internal_bonds = [idx for idx in tn_dict.keys() if len(tn_dict[idx]) == 2]
        return internal_bonds

    def get_external_indices(self) -> List[str]:
        """
        Get the external bonds of the tensor network.

        Returns:
            Indices that are connected to 1 tensor in the network.
        """
        tn_dict = self.get_index_to_tensor_dict()
        external_bonds = [idx for idx in tn_dict.keys() if len(tn_dict[idx]) == 1]
        return external_bonds

    def get_all_indices(self) -> List[str]:
        """
        Get all indices in the tensor network.

        Returns:
            A list of all index names.
        """
        dict = self.get_index_to_tensor_dict()
        return list(dict.keys())

    def get_all_labels(self) -> List[str]:
        """
        Get all labels in the tensor network.

        Returns:
            A list of all label names.
        """
        all_labels = []
        for t in self.tensors:
            for l in t.labels:
                all_labels.append(l)
        return list(set(all_labels))

    def get_new_label(self, tensor_prefix: str = "TN_T") -> str:
        """
        Get a new tensor label with the specified prefix.

        Args:
            tensor_prefix (optional): Defaults to "TN_T".

        Returns:
            A new label that doesn't already appear in the network starting with tensor_prefix.
        """
        all_labels = [x for x in self.get_all_labels() if len(x) > len(tensor_prefix)]

        current_vals = []
        for label in all_labels:
            if (
                label[: len(tensor_prefix)] == tensor_prefix
                and label[len(tensor_prefix) :].isdigit()
            ):
                current_vals.append(int(label[len(tensor_prefix) :]))
        if len(current_vals) > 0:
            max_current_val = max(current_vals)
        else:
            max_current_val = 0
        new_label = tensor_prefix + str(max_current_val + 1)

        return new_label

    def get_tensors_from_index_name(self, idx: str) -> List[Tensor]:
        """
        Get all tensors connected to a given index.

        Args:
            idx: The index name.

        Returns
            A list of tensors with idx as one of their indices.
        """
        return [t for t in self.tensors if idx in t.indices]

    def get_tensors_from_label(self, label: str) -> List[Tensor]:
        """
        Get all tensors connected to a given label.

        Args:
            label: The label name.

        Returns
            A list of tensors with label as one of their labels.
        """
        return [t for t in self.tensors if label in t.labels]

    def contract_index(self, idx: str) -> None:
        """
        Contract an index in the tensor network.

        Args:
            idx: The name of the index to contract.
        """
        tensors = self.get_tensors_from_index_name(idx)

        array0, array1 = tensors[0].data, tensors[1].data
        indices0, indices1 = tensors[0].indices, tensors[1].indices

        output_indices = [i for i in indices0 if i != idx] + [
            i for i in indices1 if i != idx
        ]
        new_data = ctg.array_contract(
            arrays=[array0, array1],
            inputs=[indices0, indices1],
            output=output_indices,
            cache_expression=True,
            prefer_einsum=True,
        )
        new_labels = [self.get_new_label()]
        if len(new_data.shape) > len(output_indices):
            new_data = new_data.reshape(new_data.shape[1:])
        new_tensor = Tensor(new_data, output_indices, new_labels)

        pos = self.tensors.index(tensors[0])
        self.tensors.remove(tensors[0])
        self.tensors.remove(tensors[1])
        self.indices.remove(idx)
        self.add_tensor(new_tensor, position=pos)

        return

    def compress_index(
        self,
        idx: str,
        max_bond: int,
        reverse_direction: bool = False,
        tol: float = 1e-12,
    ) -> None:
        """
        Compress a given index using SVD.

        Args:
            idx: The index to compress.
            max_bond: The maximum bond dimension for this index.
            reverse_direction: If True the second tensor will become an isometry
            tol: Tolerance for discarding singular values
        """
        if reverse_direction:
            tensors = self.get_tensors_from_index_name(idx)[::-1]
        else:
            tensors = self.get_tensors_from_index_name(idx)

        array0, array1 = tensors[0].data, tensors[1].data
        indices0, indices1 = tensors[0].indices, tensors[1].indices
        dims0, dims1 = tensors[0].dimensions, tensors[1].dimensions

        output_indices = [i for i in indices0 if i != idx] + [
            i for i in indices1 if i != idx
        ]

        new_data = ctg.array_contract(
            arrays=[array0, array1],
            inputs=[indices0, indices1],
            output=output_indices,
            cache_expression=True,
            prefer_einsum=True,
        )
        temp_tensor = Tensor(new_data, output_indices, ["TEMP"])

        input_idxs = [i for i in indices0 if i != idx]
        output_idxs = [i for i in indices1 if i != idx]
        temp_tensor.tensor_to_matrix(input_idxs, output_idxs)

        bond_dim = min([max_bond, temp_tensor.data.shape[0], temp_tensor.data.shape[1]])
        if bond_dim >= min([temp_tensor.data.shape[0], temp_tensor.data.shape[1]]) - 1:
            u, s, vh = svd(temp_tensor.data.todense(), full_matrices=False)
        else:
            u, s, vh = svds(temp_tensor.data, k=bond_dim)

        s = s[s > 1e-14]
        sq = s**2
        cumulative = np.cumsum(sq[::-1])[::-1]
        keep_dim = len(s)
        for k in range(len(s)):
            if cumulative[k] < tol**2:
                keep_dim = k + 1
                break
        keep_dim = min(keep_dim, bond_dim)

        new_data0 = sparse.COO.from_numpy(vh[:keep_dim, :])
        new_data1 = sparse.COO.from_numpy(u[:, :keep_dim] * s[:keep_dim])

        idx_pos0 = indices0.index(idx)
        idx_pos1 = indices1.index(idx)
        new_dims0 = (keep_dim,) + dims0[:idx_pos0] + dims0[idx_pos0 + 1 :]
        new_dims1 = dims1[:idx_pos1] + dims1[idx_pos1 + 1 :] + (keep_dim,)
        new_indices0 = [idx] + indices0[:idx_pos0] + indices0[idx_pos0 + 1 :]
        new_indices1 = indices1[:idx_pos1] + indices1[idx_pos1 + 1 :] + [idx]

        new_data0 = new_data0.reshape(new_dims0)
        new_data1 = new_data1.reshape(new_dims1)

        tensors[0].data = new_data0
        tensors[1].data = new_data1
        tensors[0].indices = new_indices0
        tensors[1].indices = new_indices1
        tensors[0].dimensions = new_dims0
        tensors[1].dimensions = new_dims1

        tensors[0].reorder_indices(indices0)
        tensors[1].reorder_indices(indices1)

        return

    def pop_tensors_by_label(self, labels: List[str]) -> List[Tensor]:
        """
        Remove tensors from the network given a set of labels.

        Args:
            labels: The list of labels to search for.

        Returns:
            The (possibly empty) list of removed tensors.
        """
        tensors = []
        for tensor in self.tensors:
            has_all_labels = True
            for label in labels:
                if label not in tensor.labels:
                    has_all_labels = False
            if has_all_labels:
                tensors.append(tensor)

        tn_dict = self.get_index_to_tensor_dict()
        for t in tensors:
            self.tensors.remove(t)

        for idx in tn_dict:
            still_exists = False
            if len(tn_dict[idx]) > 0:
                still_exists = True
            if not still_exists:
                self.indices.remove(idx)

        return tensors

    def add_tensor(
        self, tensor: Tensor, position: int = None, add_label: bool = False
    ) -> None:
        """
        Add a tensor to the network.

        Args:
            tensor: The tensor to add.
        """
        if add_label:
            unique_label = self.get_new_label("TN_T")
            tensor.labels.append(unique_label)
        if position is None:
            self.tensors.append(tensor)
        else:
            self.tensors.insert(position, tensor)
        for idx in tensor.indices:
            if idx not in self.indices:
                self.indices.append(idx)
        return

    def contract_entire_network(self) -> Union[Tensor, complex]:
        """
        Contracts all internal indices in the network.

        Returns:
            A tensor whose indices were the external indices of the network, or a float if there were no external indices.
        """
        output_indices = self.get_external_indices()
        output_labels = [self.get_new_label("TN_T")]
        arrays = [t.data for t in self.tensors]
        input_indices = [t.indices for t in self.tensors]

        output_tensor_data = ctg.array_contract(
            arrays=arrays,
            inputs=input_indices,
            output=output_indices,
            cache_expression=True,
            prefer_einsum=True,
        )
        if len(output_indices) == 0:
            return complex(output_tensor_data.flatten()[0])
        else:
            output_tensor = Tensor(output_tensor_data, output_indices, output_labels)
            return output_tensor

    def compute_environment_tensor_by_label(
        self, labels: List[str], replace_tensor: bool = False
    ) -> Union[Tensor, None]:
        """
        Compute the environment of a tensor in the network given a set of labels.

        Args:
            labels: The labels to look for.
            replace_tensor (optional): When True replaces the original tensor in the network by its environment. Default is False.

        Returns:
            If replace_tensor is True, return type is None. Otherwise, returns the environment tensor.
        """
        popped_tensor = self.pop_tensors_by_label(labels)[0]
        output_tensor = self.contract_entire_network()

        if replace_tensor:
            self = TensorNetwork([output_tensor])
        else:
            self.add_tensor(popped_tensor)
            return output_tensor

        return

    def new_index_name(
        self, index_prefix: str = "B", num_new_indices: int = 1
    ) -> Union[str, List[str]]:
        """
        Generate a new index name not already in use.

        Args:
            index_prefix (optional): Default is "B".
            num_new_indices (optional): Number of new names required. Default is 1.

        Returns:
            The new index name. Returned as a str if num_new_indices=1, otherwise returned as List[str].
        """
        current_indices = [x for x in self.indices if len(x) > len(index_prefix)]
        current_vals = []
        for idx in current_indices:
            if (
                idx[: len(index_prefix)] == index_prefix
                and idx[len(index_prefix) :].isdigit()
            ):
                current_vals.append(int(idx[len(index_prefix) :]))
        if len(current_vals) > 0:
            max_current_val = max(current_vals)
        else:
            max_current_val = 0
        new_indices = [
            index_prefix + str(max_current_val + i)
            for i in range(1, num_new_indices + 1)
        ]

        if num_new_indices == 1:
            return new_indices[0]
        return new_indices

    def combine_indices(
        self, idxs: List[str], new_index_name: str | None = None
    ) -> None:
        """
        Combine two or more indices within the network. Only valid when all indices are between the same two tensors.

        Args:
            idxs: The indices to combine.
            new_index_name (optional): What to call the resulting combined index.
        """
        tensors = self.get_tensors_from_index_name(idxs[0])
        for t in tensors:
            t.combine_indices(idxs, new_index_name)
        self.indices = self.get_all_indices()
        return

    def svd(
        self,
        tensor: Tensor,
        input_indices: List[str],
        output_indices: List[str],
        max_bond: int | None = None,
        new_index_name: str | None = None,
        new_labels: List[List[str]] | None = None,
        tol: float | None = 1e-12,
    ) -> None:
        """
        Perform an SVD on a tensor.

        Args:
            tensor: The tensor.
            input_indices: Indices to be treated as one side of SVD.
            output_indices: Indices to be treated as other side of SVD.
            max_bond: The maximum bond dimension allwoed.
            new_index_name (optional): What to call the resulting new index.
            tol: Tolerance for keeping singular values
        """
        original_position = self.tensors.index(tensor)
        original_labels = tensor.labels
        original_input_dims = [tensor.get_dimension_of_index(x) for x in input_indices]
        original_output_dims = [
            tensor.get_dimension_of_index(x) for x in output_indices
        ]
        tensor.tensor_to_matrix(input_indices, output_indices)
        if not max_bond:
            max_bond = min([tensor.dimensions[0], tensor.dimensions[1]])
        else:
            max_bond = min([max_bond, tensor.dimensions[0], tensor.dimensions[1]])

        if max_bond >= min([tensor.dimensions[0], tensor.dimensions[1]]) - 1:
            u, s, vh = svd(tensor.data.todense(), full_matrices=False)
        else:
            u, s, vh = svds(tensor.data, k=max_bond)

        s = s[s > 1e-14]
        sq = s**2
        cumulative = np.cumsum(sq[::-1])[::-1]
        keep_dim = len(s)
        for k in range(len(s)):
            if cumulative[k] < tol**2:
                keep_dim = k + 1
                break
        keep_dim = min(keep_dim, max_bond)

        threshold = 1e-14
        data0 = vh[:keep_dim, :]
        data0[np.abs(data0) < threshold] = 0.0
        data1 = u[:, :keep_dim] * s[:keep_dim]
        data1[np.abs(data1) < threshold] = 0.0

        tensor0_data = sparse.COO.from_numpy(data0)
        tensor1_data = sparse.COO.from_numpy(data1)

        if not new_index_name:
            new_index_name = self.new_index_name()

        tensor0_labels = [self.get_new_label("TN_T")]
        tensor1_labels = [self.get_new_label("TN_T")]
        if new_labels:
            tensor0_labels = tensor0_labels + new_labels[0]
            tensor1_labels = tensor1_labels + new_labels[1]

        tensor0_indices = [new_index_name] + input_indices
        tensor1_indices = output_indices + [new_index_name]

        tensor0_dims = [keep_dim] + original_input_dims
        tensor1_dims = original_output_dims + [keep_dim]

        tensor0_data = sparse.reshape(tensor0_data, tensor0_dims)
        tensor1_data = sparse.reshape(tensor1_data, tensor1_dims)

        tensor0 = Tensor(tensor0_data, tensor0_indices, tensor0_labels)
        tensor1 = Tensor(tensor1_data, tensor1_indices, tensor1_labels)

        self.pop_tensors_by_label(original_labels)
        self.add_tensor(tensor0, original_position)
        self.add_tensor(tensor1, original_position + 1)
        return

    def compress(self, max_bond: int) -> None:
        """
        Compress the tensor network using SVD.

        Args:
            max_bond: The maximum bond dimension allowed.
        """
        internal_indices = self.get_internal_indices()
        for index in internal_indices:
            self.compress_index(index, max_bond)
        return

    def draw(
        self,
        node_size: int | None = None,
        x_len: int | None = None,
        y_len: int | None = None,
    ):
        """
        Visualise tensor network.

        Args:
            node_size: Size of nodes in figure (optional)
            x_len: Figure width (optional)
            y_len: Figure height (optional)

        Returns:
            Displays plot.
        """
        if self.name == "QuantumCircuit":
            draw_quantum_circuit(self.tensors, node_size, x_len, y_len)

        else:
            draw_arbitrary_tn(self.tensors)



import copy
from typing import Tuple

import numpy as np
import sparse
from numpy import ndarray
from scipy.sparse.linalg import eigs
from sparse import SparseArray

from .mpo import MatrixProductOperator
from .mps import MatrixProductState
from .tensor import Tensor
from .tn import TensorNetwork


class DMRG:
    def __init__(
        self,
        hamiltonian: dict[str, complex]
        | tuple[ndarray, ndarray, float]
        | MatrixProductOperator,
        max_mps_bond: int,
        method: str = "two-site",
        convergence_threshold: float = 1e-9,
        initial_state: MatrixProductState | None = None,
    ) -> "DMRG":
        """
        Constructor for the DMRG class.

        Args:
            hamiltonian: A dict of the form {pauli_string : weight} or a tuple of (one_e_integrals, two_e_integrals, nuc_energy) or an MPO
            max_mpo_bond: The maximum bond to use for the Hamiltonian MPO construction.
            max_mps_bond: The maximum bond to use for MPS during DMRG.
            method: Which method to use. One of "one-site", and "two-site". Defaults to "one-site".
            convergence_threshold: DMRG terminates once two successive sweeps differ in energy by less than this value.
            initial_state: The starting point for the DMRG calculation. If not provided, random MPS is generated.

        Returns:
            The DMRG object.
        """
        self.hamiltonian = hamiltonian
        self.method = method
        self.convergence_threshold = convergence_threshold
        if isinstance(hamiltonian, dict):
            self.num_sites = len(list(hamiltonian.keys())[0])
            self.hamiltonian_type = "qubit"
        elif isinstance(hamiltonian, MatrixProductOperator):
            self.num_sites = hamiltonian.num_sites
            self.hamiltonian_type = "MPO"
        else:
            self.num_sites = len(hamiltonian[0])
            self.nuc_energy = hamiltonian[2]
            self.hamiltonian_type = "fermionic"
        self.max_mps_bond = max_mps_bond
        self.current_max_mps_bond = 2
        self.mps = self.set_initial_state(initial_state)
        if self.hamiltonian_type == "MPO":
            self.mpo = self.add_trivial_tensors_mpo(hamiltonian)
        else:
            self.mpo = self.set_hamiltonian_mpo()
        self.energy = self.set_initial_energy()
        self.all_energies = [self.energy]
        self.left_block_cache = []
        self.right_block_cache = []
        self.left_block, self.right_block = self.initialise_blocks()

        return

    def set_initial_state(
        self, input: MatrixProductState | None = None
    ) -> MatrixProductState:
        """
        Set the initial state for DMRG.

        Args:
            input: Either a given MPS or a string indicating a construction method, one of "random" or "HF"
        """
        match input:
            case None:
                mps = MatrixProductState.random_quantum_state_mps(
                    self.num_sites, self.current_max_mps_bond
                )
            case _:
                if isinstance(input, MatrixProductState):
                    mps = input
                else:
                    raise ValueError("Given initial_state not supported")

        mps = self.add_trivial_tensors_mps(mps)

        return mps

    def set_hamiltonian_mpo(self) -> MatrixProductOperator:
        """
        Convert the Hamiltonian to an MPO for DMRG.
        """

        match self.hamiltonian_type:
            case "qubit":
                mpo = MatrixProductOperator.from_hamiltonian(self.hamiltonian, np.inf)
            case "fermionic":
                mpo = MatrixProductOperator.from_electron_integral_arrays(
                    self.hamiltonian[0], self.hamiltonian[1]
                )

        mpo = self.add_trivial_tensors_mpo(mpo)

        return mpo

    def set_initial_energy(self) -> float:
        mps = copy.deepcopy(self.mps)
        mpo = copy.deepcopy(self.mpo)
        return mps.compute_expectation_value(mpo)

    def add_trivial_tensors_mps(self, mps: MatrixProductState) -> MatrixProductState:
        """
        Add trivial tensors to MPS.
        """
        mps.tensors[0].data = sparse.reshape(
            mps.tensors[0].data, (1,) + mps.tensors[0].dimensions
        )
        mps.tensors[-1].data = sparse.reshape(
            mps.tensors[-1].data,
            (mps.tensors[-1].dimensions[0], 1, mps.tensors[-1].dimensions[1]),
        )

        trivial_array = sparse.COO.from_numpy(
            np.array([1], dtype=complex).reshape(1, 1)
        )
        all_arrays = (
            [trivial_array]
            + [mps.tensors[i].data for i in range(self.num_sites)]
            + [trivial_array]
        )
        mps = MatrixProductState.from_arrays(all_arrays)
        return mps

    def add_trivial_tensors_mpo(
        self, mpo: MatrixProductOperator
    ) -> MatrixProductOperator:
        """
        Add trivial tensors to MPO.
        """
        mpo.tensors[0].data = sparse.reshape(
            mpo.tensors[0].data, (1,) + mpo.tensors[0].dimensions
        )
        mpo.tensors[-1].data = sparse.reshape(
            mpo.tensors[-1].data,
            (
                mpo.tensors[-1].dimensions[0],
                1,
                mpo.tensors[-1].dimensions[1],
                mpo.tensors[-1].dimensions[2],
            ),
        )

        trivial_array = sparse.COO.from_numpy(
            np.array([1], dtype=complex).reshape(1, 1, 1)
        )
        all_arrays = (
            [trivial_array]
            + [mpo.tensors[i].data for i in range(self.num_sites)]
            + [trivial_array]
        )
        mpo = MatrixProductOperator.from_arrays(all_arrays)
        return mpo

    def remove_trivial_tensors_mps(self, mps: MatrixProductState) -> MatrixProductState:
        """
        Remove trivial tensors from MPS.
        """
        zero_array = sparse.COO.from_numpy(
            np.array([1], dtype=complex).reshape(
                1,
            )
        )
        zero_tensor_top = Tensor(zero_array, ["P1"], ["ZERO"])
        zero_tensor_bottom = Tensor(zero_array, [f"P{self.num_sites+2}"], ["ZERO"])
        self.mps.add_tensor(zero_tensor_top)
        self.mps.add_tensor(zero_tensor_bottom)
        self.mps.contract_index("P1")
        self.mps.contract_index(f"P{self.num_sites+2}")
        self.mps.contract_index("B1")
        self.mps.contract_index(f"B{self.num_sites+1}")
        arrays = []
        for idx in range(2, self.num_sites + 2):
            arrays.append(self.mps.get_tensors_from_index_name(f"P{idx}")[0].data)
        mps = MatrixProductState.from_arrays(arrays)
        # mps.multiply_by_constant(2)

        return mps

    def remove_trivial_tensors_mpo(
        self, mpo: MatrixProductOperator
    ) -> MatrixProductOperator:
        """
        Remove trivial tensors from MPS.
        """
        zero_array = sparse.COO.from_numpy(
            np.array([1], dtype=complex).reshape(
                1,
            )
        )
        zero_tensor_top_right = Tensor(zero_array, ["R1"], ["ZERO"])
        zero_tensor_bottom_right = Tensor(
            zero_array, [f"R{self.num_sites+2}"], ["ZERO"]
        )
        zero_tensor_top_left = Tensor(zero_array, ["L1"], ["ZERO"])
        zero_tensor_bottom_left = Tensor(zero_array, [f"L{self.num_sites+2}"], ["ZERO"])
        self.mpo.add_tensor(zero_tensor_top_right)
        self.mpo.add_tensor(zero_tensor_bottom_right)
        self.mpo.add_tensor(zero_tensor_top_left)
        self.mpo.add_tensor(zero_tensor_bottom_left)
        self.mpo.contract_index("R1")
        self.mpo.contract_index(f"R{self.num_sites+2}")
        self.mpo.contract_index("L1")
        self.mpo.contract_index(f"L{self.num_sites+2}")
        self.mpo.contract_index("B1")
        self.mpo.contract_index(f"B{self.num_sites+1}")
        arrays = []
        for idx in range(2, self.num_sites + 2):
            arrays.append(self.mpo.get_tensors_from_index_name(f"R{idx}")[0].data)
        mpo = MatrixProductOperator.from_arrays(arrays)
        # mpo.multiply_by_constant(4)

        return mpo

    def initialise_left_block(self) -> Tensor:
        """
        Initialise the left block of the DMRG routine.
        """
        dag = copy.deepcopy(self.mps.tensors[0].data.conj())
        ham = copy.deepcopy(self.mpo.tensors[0].data)
        mps = copy.deepcopy(self.mps.tensors[0].data)

        dag_tensor = Tensor(dag, ["Ldag", "TEMP1"], ["DAG"])
        ham_tensor = Tensor(ham, ["Lham", "TEMP1", "TEMP2"], ["HAM"])
        mps_tensor = Tensor(mps, ["Lmps", "TEMP2"], ["MPS"])

        tn = TensorNetwork([dag_tensor, ham_tensor, mps_tensor])
        left_block = tn.contract_entire_network()
        left_block.reorder_indices(["Ldag", "Lham", "Lmps"])
        left_block.labels.append("LEFT_OF_SITE_1")

        return left_block

    def initialise_blocks(self) -> Tuple[Tensor]:
        """
        Initialise the left and right blocks of the DMRG routine.

        Returns:
            A tuple of the initial left block and the initial right block.
        """
        # Set up the rightmost block
        dag = copy.deepcopy(self.mps.tensors[-1].data.conj())
        ham = copy.deepcopy(self.mpo.tensors[-1].data)
        mps = copy.deepcopy(self.mps.tensors[-1].data)
        dag_tensor = Tensor(dag, ["Rdag", "TEMP1"], ["DAG"])
        ham_tensor = Tensor(ham, ["Rham", "TEMP1", "TEMP2"], ["HAM"])
        mps_tensor = Tensor(mps, ["Rmps", "TEMP2"], ["MPS"])
        tn = TensorNetwork([dag_tensor, ham_tensor, mps_tensor])
        right_block = tn.contract_entire_network()
        right_block.reorder_indices(["Rdag", "Rham", "Rmps"])
        right_block.labels.append(f"RIGHT_OF_SITE_{self.num_sites}")
        self.right_block_cache.append(right_block)

        for i in list(range(2, self.num_sites + 1))[::-1]:
            self.mps.move_orthogonality_centre(i, i + 1)
            dag = copy.deepcopy(self.mps.tensors[i].data.conj())
            ham = copy.deepcopy(self.mpo.tensors[i].data)
            mps = copy.deepcopy(self.mps.tensors[i].data)
            temp_right_block = copy.deepcopy(right_block)
            temp_right_block.indices = ["TEMP1", "TEMP3", "TEMP5"]
            dag_tensor = Tensor(dag, ["Rdag", "TEMP1", "TEMP2"], ["DAG"])
            ham_tensor = Tensor(ham, ["Rham", "TEMP3", "TEMP2", "TEMP4"], ["HAM"])
            mps_tensor = Tensor(mps, ["Rmps", "TEMP5", "TEMP4"], ["MPS"])
            tn = TensorNetwork([temp_right_block, dag_tensor, ham_tensor, mps_tensor])
            right_block = tn.contract_entire_network()
            right_block.reorder_indices(["Rdag", "Rham", "Rmps"])
            right_block.labels.append(f"RIGHT_OF_SITE_{i-1}")
            self.right_block_cache.append(right_block)

        left_block = self.initialise_left_block()
        self.right_block_cache.pop()

        if self.method == "two-site":
            right_block = copy.deepcopy(self.right_block_cache[-1])
            self.right_block_cache.pop()

        return left_block, right_block

    def update_blocks_left_sweep(self) -> None:
        """
        Update the blocks after each local optimisation.
        """
        current_site = len(self.left_block_cache) + 1
        self.left_block_cache.append(self.left_block)
        self.mps.move_orthogonality_centre(current_site + 2, current_site + 1)
        left_block = copy.deepcopy(self.left_block)

        mps = copy.deepcopy(self.mps.tensors[current_site].data)
        ham = copy.deepcopy(self.mpo.tensors[current_site].data)
        dag = copy.deepcopy(self.mps.tensors[current_site].data.conj())

        left_block.indices = ["TEMP1", "TEMP2", "TEMP3"]
        mps_tensor = Tensor(mps, ["TEMP3", "Lmps", "TEMP5"], ["MPS"])
        ham_tensor = Tensor(ham, ["TEMP2", "Lham", "TEMP4", "TEMP5"], ["HAM"])
        dag_tensor = Tensor(dag, ["TEMP1", "Ldag", "TEMP4"], ["DAG"])

        tn = TensorNetwork([left_block, mps_tensor, ham_tensor, dag_tensor])
        self.left_block = tn.contract_entire_network()
        self.left_block.reorder_indices(["Ldag", "Lham", "Lmps"])
        self.left_block.labels.append(f"LEFT_OF_SITE_{current_site+1}")

        self.right_block = copy.deepcopy(self.right_block_cache[-1])
        self.right_block_cache.pop()

        return

    def update_blocks_right_sweep(self) -> None:
        """
        Update the blocks after each local optimisation.
        """
        current_site = self.num_sites - len(self.right_block_cache)
        self.right_block_cache.append(self.right_block)
        self.mps.move_orthogonality_centre(current_site, current_site + 1)
        right_block = copy.deepcopy(self.right_block)

        mps = copy.deepcopy(self.mps.tensors[current_site].data)
        ham = copy.deepcopy(self.mpo.tensors[current_site].data)
        dag = copy.deepcopy(self.mps.tensors[current_site].data.conj())

        right_block.indices = ["TEMP1", "TEMP2", "TEMP3"]
        mps_tensor = Tensor(mps, ["Rmps", "TEMP3", "TEMP5"], ["MPS"])
        ham_tensor = Tensor(ham, ["Rham", "TEMP2", "TEMP4", "TEMP5"], ["HAM"])
        dag_tensor = Tensor(dag, ["Rdag", "TEMP1", "TEMP4"], ["DAG"])

        tn = TensorNetwork([right_block, mps_tensor, ham_tensor, dag_tensor])
        self.right_block = tn.contract_entire_network()
        self.right_block.reorder_indices(["Rdag", "Rham", "Rmps"])
        self.right_block.labels.append(f"RIGHT_OF_SITE_{current_site-1}")

        self.left_block = copy.deepcopy(self.left_block_cache[-1])
        self.left_block_cache.pop()

        return

    def combine_neighbouring_sites(self) -> SparseArray:
        """
        For the two_site method, combine neighbouring Hamiltonian sites.
        """
        current_site = len(self.left_block_cache) + 1
        next_site = current_site + 1
        ham1 = copy.deepcopy(self.mpo.tensors[current_site])
        ham2 = copy.deepcopy(self.mpo.tensors[next_site])

        tn = TensorNetwork([ham1, ham2])
        combined = tn.contract_entire_network()
        combined.combine_indices([f"R{current_site+1}", f"R{current_site+2}"], "R")
        combined.combine_indices([f"L{current_site+1}", f"L{current_site+2}"], "L")
        combined.reorder_indices([f"B{current_site}", f"B{current_site+2}", "R", "L"])

        return combined.data

    def construct_effective_matrix(
        self, current_site_mat: SparseArray | None = None
    ) -> SparseArray:
        """
        Construct the effective matrix for a step of DMRG.

        Args:
            current_site_mat (optional): The MPO tensor at the site(s) to be optimised. Defaults to single site tensor.
        """
        left_block = copy.deepcopy(self.left_block)
        right_block = copy.deepcopy(self.right_block)
        current_site = len(self.left_block_cache) + 1
        if current_site_mat is None:
            current_site_mat = copy.deepcopy(self.mpo.tensors[current_site].data)

        ham_tensor = Tensor(current_site_mat, ["Lham", "Rham", "pdag", "p"], ["HAM"])
        tn = TensorNetwork([ham_tensor, right_block, left_block])
        tensor = tn.contract_entire_network()
        tensor.reorder_indices(["Ldag", "pdag", "Rdag", "Lmps", "p", "Rmps"])
        tensor.indices = ["udag", "pdag", "ddag", "u", "p", "d"]
        tensor.tensor_to_matrix(["u", "p", "d"], ["udag", "pdag", "ddag"])

        return tensor.data

    def optimise_local_tensor(self) -> None:
        """
        Optimise the local tensor at the current site.

        Args:
            site: The site index.
        """
        site = len(self.left_block_cache) + 1
        if self.method == "two-site":
            original_dims = (
                self.mps.tensors[site].dimensions[0],
                self.mps.tensors[site + 1].dimensions[1],
                self.mps.tensors[site].dimensions[2],
                self.mps.tensors[site + 1].dimensions[2],
            )
            ham_mat = self.combine_neighbouring_sites()
            effective_matrix = self.construct_effective_matrix(ham_mat)
        else:
            original_dims = self.mps.tensors[site].dimensions + (1,)
            effective_matrix = self.construct_effective_matrix()
        w, v = eigs(effective_matrix, k=1, which="SR")
        eigval = w[0]
        eigvec = sparse.COO.from_numpy(
            v[:, 0]
        )  # This is the new optimal value at site i

        new_data = sparse.reshape(
            eigvec,
            (original_dims[0], original_dims[2] * original_dims[3], original_dims[1]),
        )
        new_data = sparse.moveaxis(new_data, [0, 1, 2], [0, 2, 1])

        if self.method == "two-site":
            new_data = sparse.moveaxis(new_data, [0, 1, 2], [2, 1, 0])
            new_data = sparse.reshape(
                new_data,
                (
                    original_dims[2],
                    original_dims[3],
                    original_dims[1],
                    original_dims[0],
                ),
            )
            new_data = sparse.moveaxis(new_data, [0, 1, 2, 3], [2, 3, 1, 0])
            temp_t = Tensor(new_data, ["u", "d", "p1", "p2"], ["TEMP"])
            temp_tn = TensorNetwork([temp_t])
            temp_tn.svd(
                temp_t,
                ["u", "p1"],
                ["d", "p2"],
                max_bond=self.current_max_mps_bond,
                new_index_name="b",
            )

            t1 = temp_tn.tensors[0]
            t1.reorder_indices(["u", "b", "p1"])
            indices1 = self.mps.tensors[site].indices
            labels1 = self.mps.tensors[site].labels
            t1.indices = indices1
            t1.labels = labels1

            t2 = temp_tn.tensors[1]
            t2.reorder_indices(["b", "d", "p2"])
            indices2 = self.mps.tensors[site + 1].indices
            labels2 = self.mps.tensors[site + 1].labels
            t2.indices = indices2
            t2.labels = labels2

            self.mps.pop_tensors_by_label(labels1)
            self.mps.add_tensor(t1, site)
            self.mps.pop_tensors_by_label(labels2)
            self.mps.add_tensor(t2, site + 1)

        else:
            original_indices = self.mps.tensors[site].indices
            original_labels = self.mps.tensors[site].labels
            self.mps.pop_tensors_by_label(original_labels)
            new_t = Tensor(new_data, original_indices, original_labels)
            self.mps.add_tensor(new_t, site)

        self.energy = eigval.real

        return

    def sweep_left_one_site(self):
        """
        Perform a left sweep.
        """
        sites = list(range(1, self.num_sites + 1))
        for site in sites:
            self.optimise_local_tensor()
            if site != self.num_sites:
                self.update_blocks_left_sweep()
        return

    def sweep_right_one_site(self):
        """
        Perform a right sweep.
        """
        sites = list(range(1, self.num_sites + 1))[::-1]
        for site in sites:
            self.optimise_local_tensor()
            if site != 1:
                self.update_blocks_right_sweep()
        return

    def sweep_left_two_site(self):
        """
        Perform a left sweep.
        """
        sites = list(range(1, self.num_sites))
        for site in sites:
            self.optimise_local_tensor()
            if site != self.num_sites - 1:
                self.update_blocks_left_sweep()
        return

    def sweep_right_two_site(self):
        """
        Perform a right sweep.
        """
        sites = list(range(1, self.num_sites))[::-1]
        for site in sites:
            self.optimise_local_tensor()
            if site != 1:
                self.update_blocks_right_sweep()
        return

    def perform_bond_expansion(self) -> None:
        """
        Expand the MPS bond dimension.
        """
        new_bond_dim = min(2 * self.current_max_mps_bond, self.max_mps_bond)
        diff = new_bond_dim - self.current_max_mps_bond
        self.current_max_mps_bond = new_bond_dim
        self.mps = self.mps.expand_bond_dimension_list(
            diff, list(range(2, self.num_sites + 1))
        )
        self.left_block_cache = []
        self.right_block_cache = []
        self.left_block, self.right_block = self.initialise_blocks()
        return

    def sub_convergence_check(self) -> None:
        """
        Check if the convergence threshold has been met for the current bond dimension.
        """
        if len(self.all_energies) < 2:
            return False
        if np.abs(self.all_energies[-1] - self.all_energies[-2]) < 1e-3:
            return True
        return False

    def convergence_check(self) -> None:
        """
        Check if the convergence threshold has been met for the max bond dimension.
        """
        if len(self.all_energies) < 2:
            return False
        if self.current_max_mps_bond == self.max_mps_bond:
            if (
                np.abs(self.all_energies[-1] - self.all_energies[-2])
                < self.convergence_threshold
            ):
                return True
        return False

    def run(self, maxiter: int) -> Tuple[float, MatrixProductState]:
        """
        Find the groundstate of an MPO with DMRG.

        Args:
            maxiter: The maximum number of DMRG sweeps.

        Returns:
            A tuple of the DMRG energy and the DMRG state.
        """
        if self.method == "one-site":
            for _ in range(maxiter):
                self.sweep_left_one_site()
                self.sweep_right_one_site()
                self.all_energies.append(self.energy)
                if self.convergence_check():
                    break
                elif self.sub_convergence_check():
                    self.perform_bond_expansion()
        elif self.method == "two-site":
            for _ in range(maxiter):
                self.sweep_left_two_site()
                self.sweep_right_two_site()
                self.all_energies.append(self.energy)
                if self.convergence_check():
                    break
                elif self.sub_convergence_check():
                    self.perform_bond_expansion()

        self.mps = self.remove_trivial_tensors_mps(self.mps)
        self.mpo = self.remove_trivial_tensors_mpo(self.mpo)

        if self.hamiltonian_type == "fermionic":
            self.energy = self.energy + self.nuc_energy

        return (self.energy, self.mps)



import copy
import re

import numpy as np
from numpy import ndarray
from qiskit import QuantumCircuit
from qiskit.circuit.library import UnitaryGate
from scipy.linalg import svd

from ..fidelity_metrics import hilbert_schmidt_fidelity
from ..mps import MatrixProductOperator
from ..tn import TensorNetwork


class MPOOptimiser:
    """
    A class for locally optimising a quantum circuit with respect to a reference MPO and the HS distance
    """

    def __init__(self, qc: QuantumCircuit, reference: MatrixProductOperator) -> None:
        """
        Constructor

        Args:
            qc: The quantum circuit that will be optimised
            reference: The reference MPO
        """
        self.qc = qc
        self.num_qubits = qc.num_qubits
        self.tn = TensorNetwork.from_qiskit_circuit(qc)
        for t in self.tn.tensors:
            t.labels.append(f"variational_site_{self.tn.tensors.index(t)+1}")
        self.reference = reference
        self.left_tn_indices, self.right_tn_indices = self.get_tn_external_indices(
            self.tn
        )
        self.error = self.calculate_error()
        self.fidelity = self.get_fidelity()
        self.optimisation_dict = {
            "optimisation_iteration": [0],
            "error": [self.error],
            "fidelity": [self.fidelity],
        }

    def get_tn_external_indices(self, tn: TensorNetwork) -> tuple[list[str], list[str]]:
        """
        Get the left and right indices of TN
        """

        def _index_splitter(idx):
            """Split the TN index into QW<number>, N<number>"""
            match = re.match(r"(QW\d+)(N\d+)", idx)
            qw, n = match.groups()
            return qw, n

        def _get_left_tn_indices():
            """Get all TN indices with N number 0"""
            left_tn_indices = [0] * self.num_qubits
            for t in tn.tensors:
                for idx in t.indices:
                    qw, n = _index_splitter(idx)
                    if n[1:] == "0":
                        left_tn_indices[int(qw[2:]) - 1] = idx
            return left_tn_indices

        def _get_right_tn_indices():
            """Get all TN indices with maximum N number for each QW number"""
            index_dict = {f"QW{x}": [] for x in range(1, self.num_qubits + 1)}
            for t in tn.tensors:
                for idx in t.indices:
                    qw, n = _index_splitter(idx)
                    index_dict[qw].append(int(n[1:]))
            right_tn_tensors = []
            for k, v in index_dict.items():
                max_n_number = max(v)
                index = k + "N" + str(max_n_number)
                right_tn_tensors.append(index)
            return right_tn_tensors

        left_tn_indices = _get_left_tn_indices()
        right_tn_indies = _get_right_tn_indices()
        return left_tn_indices, right_tn_indies

    def trace_rrdag(self) -> complex:
        """
        Calculate Tr(RR^dag) where R is the reference MPO
        """
        r1 = copy.deepcopy(self.reference)
        r2 = copy.deepcopy(self.reference)
        r2.dagger()
        mpo = r1 * r2
        tr = mpo.trace()
        return tr

    def trace_rdagt(self) -> complex:
        """
        Calculate Tr(R^dagT) where R is the reference MPO and T is the TN
        """
        r = copy.deepcopy(self.reference)
        r.dagger()
        tn = copy.deepcopy(self.tn)

        def _index_splitter(idx):
            """Split the TN index into QW<number>, N<number>"""
            match = re.match(r"(QW\d+)(N\d+)", idx)
            qw, n = match.groups()
            return qw, n

        for t in tn.tensors:
            original_t_indices = t.indices
            new_t_indices = []
            for idx in original_t_indices:
                qw, _ = _index_splitter(idx)
                if idx in self.left_tn_indices:
                    new_t_indices.append(f"T{qw[2:]}")
                elif idx in self.right_tn_indices:
                    new_t_indices.append(f"V{qw[2:]}")
                else:
                    new_t_indices.append(idx)
            t.indices = new_t_indices
        r.set_default_indices("A", "V", "T")

        full_tn = TensorNetwork(tn.tensors + r.tensors)
        tr = full_tn.contract_entire_network()
        return tr

    def trace_tdagr(self) -> complex:
        """
        Calculate Tr(TR^dag) where R is the reference MPO and T is the TN
        """
        return self.trace_rdagt().conjugate()

    def trace_ttdag(self) -> complex:
        """
        Calculate Tr(TT^dag) where T is the TN
        """
        return 2**self.num_qubits

    def calculate_error(self) -> float:
        """
        Calculate the squared Frobenius norm between the reference MPO and the TN
        """
        error = (
            self.trace_rrdag()
            - self.trace_rdagt()
            - self.trace_tdagr()
            + self.trace_ttdag()
        )
        return max(error.real, 0.0)  # It will be real anyway

    def get_fidelity(self) -> float:
        """
        Get the fidelity
        """
        tn_mpo = MatrixProductOperator.from_qiskit_circuit(self.qc)
        fid = hilbert_schmidt_fidelity(self.reference, tn_mpo)
        return fid

    def build_trace_rdagt_tn(self) -> TensorNetwork:
        """
        Build the TN for Tr(R^dagT)
        """

        def _index_splitter(idx):
            """Split the TN index into QW<number>, N<number>"""
            match = re.match(r"(QW\d+)(N\d+)", idx)
            qw, n = match.groups()
            return qw, n

        r = copy.deepcopy(self.reference)
        r.dagger()
        tn = copy.deepcopy(self.tn)

        r.set_default_indices(internal_prefix="A", input_prefix="T", output_prefix="V")
        for t in tn.tensors:
            original_t_indices = t.indices
            new_t_indices = []
            for idx in original_t_indices:
                qw, _ = _index_splitter(idx)
                if idx in self.left_tn_indices:
                    new_t_indices.append(f"V{qw[2:]}")
                elif idx in self.right_tn_indices:
                    new_t_indices.append(f"T{qw[2:]}")
                else:
                    new_t_indices.append(idx)
            t.indices = new_t_indices

        tn = TensorNetwork(tn.tensors + r.tensors)
        return tn

    def get_environment_vector(self, variational_index: int) -> ndarray:
        """
        Get the environment vector at the given site

        Args:
            variational_index: The index of the current site

        Returns:
            The vector of the environment of the current site
        """
        tn = self.build_trace_rdagt_tn()
        site_label = f"variational_site_{variational_index}"
        popped_t = tn.pop_tensors_by_label([site_label])
        env_tensor = tn.contract_entire_network()
        env_copy = copy.deepcopy(env_tensor)
        output_inds = popped_t[0].indices
        env_copy.combine_indices(output_inds)
        env_vec = env_copy.data.todense()
        return env_vec

    def get_closest_unitary(self, mat: ndarray) -> ndarray:
        """
        Get the closest unitary to a given matrix

        Args:
            mat: The input matrix

        Returns:
            The closest unitary to mat under Frobenius norm
        """
        u, _, vh = svd(mat, full_matrices=False)
        unitary_part = u @ vh
        return unitary_part

    def update_circuit(self, variational_index: int, optimal_update: ndarray) -> None:
        """
        Update the quantum circuit with the optimal local update

        Args:
            variational_index: The local index to be updated
            optimal_value: The optimal update array
        """
        new_inst = UnitaryGate(optimal_update)
        qidxs = [
            self.qc.data[variational_index - 1].qubits[x]._index
            for x in range(len(self.qc.data[variational_index - 1].qubits))
        ]
        self.qc.data[variational_index - 1] = (new_inst, qidxs[::-1], [])
        return

    def set_tn(self) -> None:
        """
        Reset TN after changes to qc
        """
        self.tn = TensorNetwork.from_qiskit_circuit(self.qc)
        for t in self.tn.tensors:
            t.labels.append(f"variational_site_{self.tn.tensors.index(t)+1}")

    def local_update(self, variational_index: int) -> None:
        """
        Perform a local update

        Args:
            variational_index: The index of the current site
        """
        local_tensor = self.tn.tensors[variational_index - 1]
        local_indices = local_tensor.indices
        local_dimensions = [
            local_tensor.get_dimension_of_index(idx) for idx in local_indices
        ]
        dim = np.prod(local_dimensions)

        env_vec = self.get_environment_vector(variational_index)
        update = env_vec
        update = update.reshape((int(np.sqrt(dim)), int(np.sqrt(dim))))
        unitary_update = self.get_closest_unitary(update)
        self.update_circuit(variational_index, unitary_update)
        self.set_tn()
        return

    def run(self, num_sweeps: int = 10) -> QuantumCircuit:
        """
        Optimise the ansatz to match the reference

        Args:
            num_sweeps: The number of sweeps to perform

        Returns:
            The optimised quantum circuit
        """
        for it_number in range(num_sweeps):
            for idx in range(1, len(self.qc.data) + 1):
                self.local_update(idx)
            for idx in list(range(1, len(self.qc.data) + 1))[::-1]:
                self.local_update(idx)
            self.error = self.calculate_error()
            self.fidelity = self.get_fidelity()
            self.optimisation_dict["optimisation_iteration"].append(it_number + 1)
            self.optimisation_dict["error"].append(self.error)
            self.optimisation_dict["fidelity"].append(self.fidelity)
        return self.qc



from abc import ABC, abstractmethod

from qiskit import QuantumCircuit

from .result import Result


class QuantumAlgorithm(ABC):
    """
    Abstract base class for quantum algorithms.
    """

    @property
    @abstractmethod
    def circuit(self) -> QuantumCircuit:
        """Get the circuit from the algorithm"""
        pass

    @abstractmethod
    def run(self, **kwargs) -> Result:
        """Run the full algorithm pipeline. Returns result object or final value."""
        pass

    # @abstractmethod
    # def construct_circuit(self, **kwargs):
    #     """Return the circuit(s) that represent the quantum part of the algorithm."""
    #     pass

    @abstractmethod
    def set_backend(self, backend, **kwargs) -> None:
        """Attach a QuantumBackend instance for execution."""
        pass

    # @abstractmethod
    # def get_result(self):
    #     """Return structured results."""
    #     pass



import json


class Result:
    """A class to store results of quantum algorithms"""

    def __init__(
        self,
        result: float | None = None,
        measurements: dict[str, int] | None = None,
        parameters: list[float] | None = None,
        metadata: dict | None = None,
        **kwargs,
    ):
        """Constructor

        Args:
            measurements: The measurement results {bitstring : count}
            result: The result of the algorithm (final energy, phase estimate, etc.)
            parameters: For variational circuits, the optimised parameters
            metadata: Timings, backend properties, etc.
        """
        self.measurements = measurements
        self.result = result
        self.parameters = parameters
        self.metadata = metadata
        self.info = kwargs

    def __str__(self):
        """Print statement"""
        output = "Quantum Alrogithm Result\n"
        if self.metadata is not None:
            if "algorithm_name" in self.metadata:
                output += f"Algorithm: {self.metadata['algorithm_name']}\n"
            if "total_runtime" in self.metadata:
                output += f"Total runtime: {self.metadata['total_runtime']}\n"
        if self.result is not None:
            output += f"Final result = {self.result}"
        return output

    def to_dict(self) -> dict:
        """Convert Result object to a dictionary"""
        return {
            "measurements": self.measurements,
            "result": self.result,
            "parameters": self.parameters,
            "metadata": self.metadata,
            "info": self.info,
        }

    def save(self, filename: str) -> None:
        """Save Result object as a json file

        Args:
            filename: The filename for the result file
        """
        with open(filename, "w") as f:
            json.dump(self.to_dict(), f, indent=4)



import json
from dataclasses import dataclass
from enum import Enum

import numpy as np
import sparse


class PauliTerm(Enum):
    """
    A class to conveniently access Pauli operators.
    """

    I = "I"
    X = "X"
    Y = "Y"
    Z = "Z"


@dataclass(frozen=True)
class UpdateValues:
    """
    A helper class for buiding Hamiltonian MPOs.
    """

    indices: tuple[int, int, int, int]
    weights: tuple[complex, complex]


def _update_array(
    array: list,
    data: list,
    weight: complex,
    p_string_idx: int,
    term: str,
    offset: bool = False,
) -> None:
    """
    A helper function to build Hamiltonian MPOs.
    """
    match term:
        case PauliTerm.I.value:
            update_values = UpdateValues((0, 0, 1, 1), (1, 1))
        case PauliTerm.X.value:
            update_values = UpdateValues((0, 1, 1, 0), (1, 1))
        case PauliTerm.Y.value:
            update_values = UpdateValues((0, 1, 1, 0), (-1j, 1j))
        case PauliTerm.Z.value:
            update_values = UpdateValues((0, 0, 1, 1), (1, -1))

    for i in [0, 1]:
        array[0].append(p_string_idx)
        if offset:
            array[1].append(p_string_idx)

        array[1 + int(offset)].append(update_values.indices[2 * i])
        array[2 + int(offset)].append(update_values.indices[(2 * i) + 1])
        data.append(update_values.weights[i] * weight)


@dataclass(frozen=True)
class UpdateValuesFermion:
    """
    A helper class for buiding Hamiltonian MPOs.
    """

    indices: tuple[int, int]
    weights: tuple[complex]


def _update_array_fermion(
    array: list,
    data: list,
    weight: complex,
    string_idx: int,
    term: str,
    offset: bool = False,
) -> None:
    """
    A helper function to build Hamiltonian MPOs.
    """

    I = np.array([[1, 0], [0, 1]], dtype=complex)
    Z = np.array([[1, 0], [0, -1]], dtype=complex)
    p = np.array([[0, 0], [1, 0]], dtype=complex)
    m = np.array([[0, 1], [0, 0]], dtype=complex)

    total_op = I.copy()
    for x in term:
        if x == "Z":
            total_op = total_op @ Z
        if x == "+":
            total_op = total_op @ p
        if x == "-":
            total_op = total_op @ m

    non_zero_vals = []
    for row in [0, 1]:
        for col in [0, 1]:
            if total_op[row, col] == 0.0:
                continue
            non_zero_vals.append((row, col, total_op[row, col]))

    if len(non_zero_vals) == 1:
        update_values = UpdateValuesFermion(
            (non_zero_vals[0][0], non_zero_vals[0][1]), (non_zero_vals[0][2],)
        )
    elif len(non_zero_vals) == 2:
        update_values = UpdateValues(
            (
                non_zero_vals[0][0],
                non_zero_vals[0][1],
                non_zero_vals[1][0],
                non_zero_vals[1][1],
            ),
            (non_zero_vals[0][2], non_zero_vals[1][2]),
        )

    for i in range(len(non_zero_vals)):
        array[0].append(string_idx)
        if offset:
            array[1].append(string_idx)

        array[1 + int(offset)].append(update_values.indices[2 * i])
        array[2 + int(offset)].append(update_values.indices[(2 * i) + 1])
        data.append(update_values.weights[i] * weight)


def array_to_dict_nonzero_indices(arr, tol=1e-10):
    """
    A helper function to build Hamiltonians.
    """
    where_nonzero = np.where(~np.isclose(arr, 0, atol=tol))
    nonzero_indices = list(zip(*where_nonzero))
    return dict(zip(nonzero_indices, arr[where_nonzero]))


class ReadMoleculeData:
    """
    A class to read information from molecule json files.
    """

    def __init__(self, filename):
        with open(filename) as f:
            data = json.load(f)

        self.geometry = data["geometry"]
        self.basis = data["basis"]
        self.num_electrons = data["electrons"]
        self.num_spatial_orbs = data["spatial_orbs"]
        self.num_spin_orbs = 2 * self.num_spatial_orbs

        self.rhf_energy = data["E_RHF"]
        self.ccsd_energy = data["E_CCSD"]
        self.ccsdpt_energy = data["E_CCSDpT"]
        self.fci_energy = data["E_FCI"]

        fci_sparse_vec = data["FCI_vector"]
        self.fci_vector = [0] * (2**self.num_spin_orbs)
        for idx, val in fci_sparse_vec.items():
            self.fci_vector[int(idx)] = val[0] + 1j * val[1]
        self.fci_vector = np.array(self.fci_vector)
        self.fci_vector_sparse = sparse.COO.from_numpy(self.fci_vector)

        self.qubit_hamiltonian = data["qubit_hamiltonian"]
        self.qubit_hamiltonian = {
            k: v[0] + 1j * v[1] for k, v in self.qubit_hamiltonian.items()
        }

        fermionic_hamiltonian = data["fermionic_hamiltonian"]
        self.one_electron_integrals = np.array(fermionic_hamiltonian[0])
        self.two_electron_integrals = np.array(fermionic_hamiltonian[1])
        self.nuclear_energy = fermionic_hamiltonian[2]
        self.fermionic_hamiltonian = (
            self.one_electron_integrals,
            self.two_electron_integrals,
            self.nuclear_energy,
        )



# Visualisation purposes
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from networkx import DiGraph

from .tensor import Tensor


def build_graph_from_tensor_network(tensors: list[Tensor]) -> DiGraph:  # type: ignore
    """
    Build a directed graph from the list of tensors and their indices.

    Args:
        tn: A TensorNetwork object.

    Returns:
        networkx.DiGraph: Directed graph representing the tensor network
    """

    G = nx.DiGraph()
    for _, tensor in enumerate(tensors):
        indices = tensor.indices
        tensor_name = [l for l in tensor.labels if l[:2] == "TN"][0]
        G.add_node(tensor_name)

        # Add edges for bottom connections and dangling indices
        for idx in indices:
            connected_tensors = [t for t in tensors if idx in t.indices]
            if len(connected_tensors) == 2:
                t1 = connected_tensors[0]
                t2 = connected_tensors[1]
                first_tensor = [l for l in t1.labels if l[:2] == "TN"][0]
                second_tensor = [l for l in t2.labels if l[:2] == "TN"][0]
                if (first_tensor, second_tensor) not in G.edges and (
                    second_tensor,
                    first_tensor,
                ) not in G.edges:
                    G.add_edge(first_tensor, second_tensor, label=idx)
            else:
                first_tensor = tensor_name
                G.add_edge(first_tensor, idx, label=idx)

    return G


def build_graph_from_MPO(tensors: list[Tensor]) -> DiGraph:  # type: ignore
    """
    Build a directed graph from the list of tensors and their indices.

    Args:
        mpo: MPO object containing tensors with indices.

    Returns:
        networkx.DiGraph: Directed graph representing the tensor network
    """
    G = nx.DiGraph()
    for i, tensor in enumerate(tensors):
        indices = tensor.indices
        tensor_name = f"Tensor_{i + 1}"
        G.add_node(tensor_name)

        # Add edges for connections and dangling indices
        for idx in indices:
            if idx.startswith("B") and i < len(tensors) - 1:
                # Connect to the next tensor
                next_tensor = f"Tensor_{i + 2}"
                G.add_edge(tensor_name, next_tensor, label=idx)
            elif idx.startswith(("R", "L")):
                # Connect dangling indices
                G.add_edge(tensor_name, idx, label=idx)
                G.add_node(idx)  # Ensure dangling index is a node
    return G


def build_graph_from_MPS(tensors: list[Tensor]) -> DiGraph:  # type: ignore
    """
    Build a directed graph from the list of tensors and their indices.

    Args:
        mps: MPS object containing tensors with indices.

    Returns:
        networkx.DiGraph: Directed graph representing the tensor network
    """
    G = nx.DiGraph()
    for i, tensor in enumerate(tensors):
        indices = tensor.indices
        tensor_name = f"Tensor_{i + 1}"
        G.add_node(tensor_name)

        # Add edges for connections and dangling indices
        for idx in indices:
            print(idx)
            if idx.startswith("B") and i < len(tensors) - 1:
                # Connect to the next tensor
                next_tensor = f"Tensor_{i + 2}"
                G.add_edge(tensor_name, next_tensor, label=idx)
            elif idx.startswith("P"):
                # Connect dangling indices
                G.add_edge(tensor_name, idx, label=idx)
                G.add_node(idx)  # Ensure dangling index is a node
    return G


def draw_quantum_circuit(
    tensors: list[Tensor],  # type: ignore
    node_size: int | None = None,
    x_len: int | None = None,
    y_len: int | None = None,
):
    """
    Visualise a tensor network representing a quantum circuit using matplotlib and networkx.

    Args:
        qc_tn: The TensorNetwork built from a quantum circuit
        node_size (int): Size of the nodes in the plot
        x_len (int): Length of the x-axis
        y_len (int): Length of the y-axis
    """
    if not x_len:
        x_len = int(np.sqrt(len(tensors))) * 5
    if not y_len:
        y_len = x_len / 2
    if not node_size:
        node_size = x_len * 5

    # Build the graph
    G = build_graph_from_tensor_network(tensors)

    # Define positions for tensors and dangling indices
    pos = {}
    vertical_spacing = 1.0
    horizontal_spacing = 1.0

    # Assign positions for tensor nodes
    nodes = [node for node in G.nodes if node.startswith("TN")]
    for node in nodes:
        tensor = [t for t in tensors if node in t.labels][0]
        tensor_labels = tensor.labels
        layer_number = int([l for l in tensor_labels if l[0] == "L"][0][1:])
        qubit_wire = int([l for l in tensor_labels if l[0] == "Q"][0][1:])
        pos[node] = (layer_number * horizontal_spacing, -qubit_wire * vertical_spacing)

    # Assign positions for dangling indices
    for edge in G.edges(data=True):
        if not edge[1].startswith("TN"):
            if edge[1] not in pos:
                if edge[1][-1] == "0":
                    pos[edge[1]] = (
                        pos[edge[0]][0] - horizontal_spacing,
                        pos[edge[0]][1],
                    )
                else:
                    pos[edge[1]] = (
                        pos[edge[0]][0] + horizontal_spacing,
                        pos[edge[0]][1],
                    )

    # Draw the graph
    plt.figure(figsize=(x_len, y_len))

    # Separate tensor and index nodes
    tensor_nodes = [node for node in G.nodes if node.startswith("TN")]

    # Draw nodes
    nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=tensor_nodes,
        node_size=node_size,
        node_color="hotpink",
        label="Tensors",
    )

    # Draw edges
    nx.draw_networkx_edges(G, pos, edge_color="gray", arrows=False)

    # Add edge labels
    edge_labels = nx.get_edge_attributes(G, "label")
    nx.draw_networkx_edge_labels(
        G, pos, edge_labels=edge_labels, font_color="red", font_size=8
    )

    # Title and axis
    plt.title("Tensor Network Visualisation", fontsize=14)
    plt.show()


def draw_mpo(
    tensors: list[Tensor],  # type: ignore
    node_size: int | None = None,
    x_len: int | None = None,
    y_len: int | None = None,
):
    """
    Visualise the MPO.

    Args:
        mpo: MPO object containing tensors with indices.
        node_size (int): Size of the nodes in the plot.
        x_len (int): Length of the x-axis.
        y_len (int): Length of the y-axis.
    """
    if not x_len:
        x_len = int(np.sqrt(len(tensors))) * 5
    if not y_len:
        y_len = x_len / 2
    if not node_size:
        node_size = x_len * 5

    # Build the graph
    G = build_graph_from_MPO(tensors)

    # Define positions for tensors and dangling indices
    pos = {}
    vertical_spacing = 1.0
    horizontal_spacing = 1.0

    # Assign positions for tensor nodes
    tensors = [node for node in G.nodes if node.startswith("Tensor")]
    for i, tensor in enumerate(tensors):
        pos[tensor] = (0, -i * vertical_spacing)

    # Assign positions for dangling indices
    for edge in G.edges(data=True):
        if edge[1].startswith("R"):
            pos[edge[1]] = (pos[edge[0]][0] + horizontal_spacing, pos[edge[0]][1])
        elif edge[1].startswith("L"):
            pos[edge[1]] = (pos[edge[0]][0] - horizontal_spacing, pos[edge[0]][1])

    # Draw the graph
    plt.figure(figsize=(x_len, y_len))

    # Separate tensor and index nodes
    tensor_nodes = [node for node in G.nodes if node.startswith("Tensor")]

    # Draw nodes
    nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=tensor_nodes,
        node_size=node_size,
        node_color="hotpink",
        label="Tensors",
    )

    # Draw edges
    nx.draw_networkx_edges(G, pos, edge_color="gray", arrows=False)

    # Add edge labels
    edge_labels = nx.get_edge_attributes(G, "label")
    nx.draw_networkx_edge_labels(
        G, pos, edge_labels=edge_labels, font_color="red", font_size=8
    )

    # Title and axis
    plt.title("MPO Visualisation", fontsize=14)
    plt.show()


def draw_mps(
    tensors: list[Tensor],  # type: ignore
    node_size: int | None = None,
    x_len: int | None = None,
    y_len: int | None = None,
):
    """
    Visualise the MPS.

    Args:
        mps: MPS object containing tensors with indices.
        node_size (int): Size of the nodes in the plot.
        x_len (int): Length of the x-axis.
        y_len (int): Length of the y-axis.
    """
    if not x_len:
        x_len = int(np.sqrt(len(tensors))) * 5
    if not y_len:
        y_len = x_len / 2
    if not node_size:
        node_size = x_len * 5

    # Build the graph
    G = build_graph_from_MPS(tensors)

    # Define positions for tensors and dangling indices
    pos = {}
    vertical_spacing = 1.0
    horizontal_spacing = 1.0

    # Assign positions for tensor nodes
    tensors = [node for node in G.nodes if node.startswith("Tensor")]
    for i, tensor in enumerate(tensors):
        pos[tensor] = (0, -i * vertical_spacing)

    # Assign positions for dangling indices
    for edge in G.edges(data=True):
        if edge[1].startswith("P"):
            pos[edge[1]] = (pos[edge[0]][0] + horizontal_spacing, pos[edge[0]][1])

    # Draw the graph
    plt.figure(figsize=(x_len, y_len))

    # Separate tensor and index nodes
    tensor_nodes = [node for node in G.nodes if node.startswith("Tensor")]

    # Draw nodes
    nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=tensor_nodes,
        node_size=node_size,
        node_color="skyblue",
        label="Tensors",
    )

    # Draw edges
    nx.draw_networkx_edges(G, pos, edge_color="gray", arrows=False)

    # Add edge labels
    edge_labels = nx.get_edge_attributes(G, "label")
    nx.draw_networkx_edge_labels(
        G, pos, edge_labels=edge_labels, font_color="red", font_size=8
    )

    # Title and axis
    plt.title("MPS Visualisation", fontsize=14)
    plt.show()


def draw_arbitrary_tn(
    tensors: list[Tensor],  # type: ignore
    node_size: int | None = None,
    x_len: int | None = None,
    y_len: int | None = None,
):
    """
    Visualise an arbitrary tensor network using a default layout from NetworkX.

    Args:
        tn: A TensorNetwork object containing tensors and indices.
        node_size (int): Size of the nodes in the plot.
        x_len (int): Length of the x-axis.
        y_len (int): Length of the y-axis.
    """
    # Build the graph from the tensor network
    G = build_graph_from_tensor_network(tensors)

    # Use default spring layout for positioning
    pos = nx.spring_layout(G)

    # Draw the graph
    plt.figure(figsize=(x_len, y_len))

    # Draw nodes
    nx.draw_networkx_nodes(
        G, pos, node_size=node_size, node_color="hotpink", label="Tensors"
    )

    # Draw edges
    nx.draw_networkx_edges(G, pos, edge_color="gray", arrows=False)

    # Add node labels
    nx.draw_networkx_labels(G, pos, font_size=10, font_color="black")

    # Add edge labels
    edge_labels = nx.get_edge_attributes(G, "label")
    nx.draw_networkx_edge_labels(
        G, pos, edge_labels=edge_labels, font_color="red", font_size=8
    )

    # Title and axis
    plt.title("Arbitrary Tensor Network Visualisation", fontsize=14)
    plt.show()



from qiskit import QuantumCircuit
from qiskit.primitives import Sampler
from qiskit_algorithms.optimizers import ADAM, COBYLA, L_BFGS_B, QNSPSA, Optimizer


def get_optimiser_callback(opt_dict: dict[str, any], index: int) -> callable:
    """
    Get callback function for optimisation.

    Args:
        opt_dict: The optimiser callback dictionary
        index: The optimisation iteration numver

    Returns:
        An optimiser callback function
    """

    def optimiser_callback(
        num_function_evals, parameters, function_value, stepsize, flag
    ):
        ind = index
        if num_function_evals and stepsize and flag:
            pass
        ind += 1
        opt_dict["optimisation_number"].append(index)
        opt_dict["optimisation_parameters"].append(parameters)
        opt_dict["optimisation_value"].append(function_value)

    return optimiser_callback


def cobyla_optimiser(
    max_iterations: int,
    convergence_threshold: float,
    opt_dict: dict[str, any],
    index: int = 0,
) -> Optimizer:
    """
    COBYLA optimiser.

    Args:
        max_iterations: The maximum number of optimisation iterations
        convergence_threshold: The convergence threshold
        opt_dict: The dictionary to pass to the optimiser callback setter
        index: The optimisation iteration number

    Returns:
        COBYLA optimiser
    """
    opt_callback = get_optimiser_callback(opt_dict=opt_dict, index=index)
    return COBYLA(
        maxiter=max_iterations, tol=convergence_threshold, callback=opt_callback
    )


def qnspsa_optimiser(
    ansatz: QuantumCircuit,
    max_iterations: int,
    opt_dict: dict[str, any],
    index: int = 0,
) -> Optimizer:
    """
    Quantum natural gradient SPSA optimiser.

    Args:
        ansatz: The ansatz circuit
        max_iterations: The maximum number of iterations
        opt_dict: The dictionary to pass to the optimiser callback setter
        index: The optimisation iteration number

    Returns:
        QNSPSA optimiser
    """
    opt_callback = get_optimiser_callback(opt_dict=opt_dict, index=index)
    fidelity = QNSPSA.get_fidelity(ansatz, sampler=Sampler())
    return QNSPSA(fidelity, maxiter=max_iterations, callback=opt_callback)


def bfgs_optimiser(
    maximum_iterations: int, opt_dict: dict[str, any], index: int = 0
) -> Optimizer:
    """
    BFGS optimiser.

    Args:
        max_iterations: The maximum number of iterations
        opt_dict: The dictionary to pass to the optimiser callback setter
        index: The optimisation iteration number

    Returns:
        BFGS optimiser
    """
    opt_callback = get_optimiser_callback(opt_dict=opt_dict, index=index)
    return L_BFGS_B(maxiter=maximum_iterations, callback=opt_callback)


def adam_optimiser(
    maximum_iterations: int, opt_dict: dict[str, any], index: int = 0
) -> Optimizer:
    """
    ADAM optimiser.

    Args:
        max_iterations: The maximum number of iterations
        opt_dict: The dictionary to pass to the optimiser callback setter
        index: The optimisation iteration number

    Returns:
        ADAM optimiser
    """
    opt_callback = get_optimiser_callback(opt_dict=opt_dict, index=index)
    return ADAM(maxiter=maximum_iterations, callback=opt_callback)



import numpy as np
from numpy import ndarray
from qiskit import QuantumCircuit
from qiskit.circuit.library import UnitaryGate
from qiskit.quantum_info import Operator
from scipy.linalg import expm


def kak_recomposition(
    sq1_params: list[float],
    sq2_params: list[float],
    cartan_params: list[float],
    sq3_params: list[float],
    sq4_params: list[float],
) -> ndarray:
    """
    Given a set of parameters, form the corresponding 2-qubit unitary U = (sq1 x sq2) * entangling * (sq3 x sq4)

    Args:
        sq1_params: Three parameters defining a single qubit gate
        sq2_params: Three parameters defining a single qubit gate
        cartan_params: Three parameters defining an entangling gate
        sq3_params: Three parameters defining a single qubit gate
        sq4_params: Three parameters defining a single qubit gate

    Returns:
        The 4x4 matrix for U
    """
    x = np.array([[0, 1], [1, 0]])
    y = np.array([[0, -1j], [1j, 0]])
    z = np.array([[1, 0], [0, -1]])
    entangling_mat = expm(
        -1j
        * (
            cartan_params[0] * np.kron(x, x)
            + cartan_params[1] * np.kron(y, y)
            + cartan_params[2] * np.kron(z, z)
        )
    )
    entangling_gate = UnitaryGate(entangling_mat)

    qc = QuantumCircuit(2)
    qc.u(sq1_params[0], sq1_params[1], sq1_params[2], 0)
    qc.u(sq2_params[0], sq2_params[1], sq2_params[2], 1)
    qc.append(entangling_gate, [1, 0])
    qc.u(sq3_params[0], sq3_params[1], sq3_params[2], 0)
    qc.u(sq4_params[0], sq4_params[1], sq4_params[2], 1)

    op = Operator(qc).reverse_qargs().data
    return op


def single_qubit_gate(params: list[float]) -> ndarray:
    """
    Get the matrix of a generic 2 qubit gate

    Args:
        params: The three paramters controlling the gate

    Returns:
        The 2x2 matrix
    """
    qc = QuantumCircuit(1)
    qc.u(params[0], params[1], params[2], 0)
    op = Operator(qc).reverse_qargs().data
    return op


def givens_rotation(rot_angle: float) -> ndarray:
    """
    Build a Givens rotation

    Args:
        rot_angle: The rotation angle

    Returns:
        The 4x4 matrix
    """
    mat = np.array(
        [
            [1, 0, 0, 0],
            [0, np.cos(rot_angle), np.sin(rot_angle), 0],
            [0, -np.sin(rot_angle), np.cos(rot_angle), 0],
            [0, 0, 0, 1],
        ],
        dtype=complex,
    )
    return mat


def symmetry_preserving_two_qubit_gate(
    subspace1_param: float, subspace2_params: list[float], subspace3_param: float
) -> ndarray:
    """
    Build a gate that preserves number and S_z symmetry

    Args:
        subspace_1_param: The parameter that controls the phase on the |00> subspace
        subspace2_params: The three parameters defining the rotation on the |01>,|10> subspace
        subspace3_param: The parameter that controls the phase on the |11> subspace

    Returns:
        The 4x4 matrix
    """
    sq_gate = single_qubit_gate(subspace2_params)
    mat = np.array(
        [
            [np.exp(1j * subspace1_param), 0, 0, 0],
            [0, sq_gate[0, 0], sq_gate[0, 1], 0],
            [0, sq_gate[1, 0], sq_gate[1, 1], 0],
            [0, 0, 0, np.exp(1j * subspace3_param)],
        ]
    )
    return mat



from timeit import default_timer

import numpy as np
from qiskit import QuantumCircuit

from ..backend.base import QuantumBackend
from ..backend.tn_backend import TNQuantumBackend
from ..base import QuantumAlgorithm
from ..result import Result
from ..utils import calculate_exp_val_stochastic, exp_pauli_string_to_circ


class QDriftSimulation(QuantumAlgorithm):
    """
    Perform Hamiltonian simulation using qDRIFT
    """

    def __init__(
        self,
        hamiltonian: dict[str, float],
        duration: float,
        error: float | None = None,
        backend: QuantumBackend | None = None,
    ) -> "QDriftSimulation":
        """
        Constructor for qDRIFT simulation class.

        Args:
            hamiltonian: The qubit Hamiltonian
            duration: The time to simulate evolution for
            num_steps: The number of Trotter steps, defaults to a sensible value
            error: The desired error
        """
        self.hamiltonian = hamiltonian
        self.norm = np.sum([np.abs(x) for x in hamiltonian.values()])
        self.duration = duration

        if error is None:
            self.error = 1e-3
        else:
            self.error = error
        self.num_terms = int(
            np.ceil(2 * (self.norm**2) * (self.duration**2) / self.error)
        )

        self._circuit = self.build_circuit(
            self.hamiltonian, self.norm, self.num_terms, self.duration
        )
        self.set_backend(backend)

    @property
    def circuit(self) -> QuantumCircuit:
        return self._circuit

    def build_circuit(
        self, ham: dict[str, float], norm: float, num_terms: int, duration: float
    ) -> QuantumCircuit:
        pauli_strings = list(ham.keys())

        num_qubits = len(pauli_strings[0])
        qc = QuantumCircuit(num_qubits)

        term_idxs = list(range(len(list(ham.keys()))))
        probs = [np.abs(weight) / norm for weight in ham.values()]
        for _ in range(num_terms):
            sample = np.random.choice(term_idxs, p=probs)
            p = list(ham.keys())[sample]
            sign = 1 if ham[p] >= 0.0 else -1
            temp_qc = exp_pauli_string_to_circ(p, norm * duration * sign / num_terms)
            qc.compose(temp_qc, inplace=True)

        return qc

    def run(
        self,
        num_shots: int = 1024,
        num_circuits: int | None = None,
        observable: dict | None = None,
    ) -> Result:
        """Run the full algorithm pipeline. Returns result object or final value."""
        start_time = default_timer()
        counts = {}
        shots_per_circ = int(num_shots / num_circuits)
        for _ in range(num_circuits):
            self._circuit = self.build_circuit(
                self.hamiltonian, self.norm, self.num_terms, self.duration
            )
            bs = self.backend.run(self.circuit, shots=shots_per_circ)
            for b, count in bs.items():
                if b in counts:
                    counts[b] += count
                else:
                    counts[b] = count
        if observable is None:
            result = None
        else:
            result = calculate_exp_val_stochastic(
                self.build_circuit,
                observable,
                self.backend,
                num_shots,
                self.hamiltonian,
                self.norm,
                self.num_terms,
                self.duration,
            )
            counts = None

        end_time = default_timer()

        metadata = {
            "algorithm_name": "Trotterisation",
            "num_shots": num_shots,
            "total_runtime": end_time - start_time,
        }
        if self.backend is not None:
            metadata["backend_name"] = self.backend.name
            metadata["backend_coupling_map"] = self.backend.coupling_map
            metadata["backend_basis_gates"] = self.backend.basis_gates
            metadata["backend_num_qubits"] = self.backend.num_qubits
        result = Result(
            result=result,
            measurements=counts,
            parameters=None,
            metadata=metadata,
        )
        return result

    def set_backend(self, backend: QuantumBackend | None = None) -> None:
        """Attach a QuantumBackend instance for execution."""
        if backend is None:
            backend = TNQuantumBackend()
        self.backend = backend
        return



from typing import Callable

import numpy as np
from numpy import ndarray

from .mpo import MatrixProductOperator
from .mps import MatrixProductState
from .qi_metrics import (
    get_all_mutual_information,
    get_one_orbital_entropy,
    get_one_orbital_rdm,
    get_two_orbital_rdm,
)


def cost_entropy(mps: MatrixProductState) -> float:
    """
    Cost based on total single-orbital entropy.
    """
    num_orbitals = mps.num_sites // 2
    return sum(get_one_orbital_entropy(mps, i + 1) for i in range(num_orbitals))


def cost_entropy_dict(num_orbitals: int) -> dict:
    s = {}
    for i in range(num_orbitals):
        s[f"S1_{i+1}"] = 1.0
    return s


def cost_total_mutual_information(mps: MatrixProductState) -> float:
    """
    Sum of all mutual informations. Lower total means less entanglement.
    """
    mi = get_all_mutual_information(mps)
    return np.sum(mi) / 2  # MI is symmetric


def cost_total_mutual_information_dict(num_orbitals: int) -> dict:
    s = {}
    for i in range(num_orbitals):
        s[f"S1_{i+1}"] = s.get(f"S1_{i+1}", 0) + 1.0
        for j in range(i + 1, num_orbitals):
            s[f"S1_{i+1}"] = s.get(f"S1_{i+1}", 0) + 1.0
            s[f"S1_{j+1}"] = s.get(f"S1_{j+1}", 0) + 1.0
            s[f"S2_{i+1}_{j+1}"] = s.get(f"S2_{i+1}_{j+1}", 0) - 1.0
    return s


def cost_mutual_info_decay(mps: MatrixProductState, decay_power: float = 2.0) -> float:
    """
    Cost function penalising long-range mutual information.
    Put highly entangled orbitals next to each other in the DMRG chain
    """
    mi = get_all_mutual_information(mps)
    n_orbs = mi.shape[0]
    cost = 0.0
    for i in range(n_orbs):
        for j in range(i + 1, n_orbs):
            distance = abs(i - j)
            cost += mi[i, j] * (distance**decay_power)
    return cost


def cost_mutual_info_decay_dict(num_orbitals: int, decay_power: float = 2.0) -> dict:
    s = {}
    for i in range(num_orbitals):
        for j in range(i + 1, num_orbitals):
            distance = abs(i - j)
            s[f"S1_{i+1}"] = s.get(f"S1_{i+1}", 0) + (distance**decay_power)
            s[f"S1_{j+1}"] = s.get(f"S1_{j+1}", 0) + (distance**decay_power)
            s[f"S2_{i+1}_{j+1}"] = s.get(f"S2_{i+1}_{j+1}", 0) - (distance**decay_power)
    return s


def cost_mutual_info_clusters(mps: MatrixProductState, threshold: float = 0.1) -> float:
    """
    Cost is the number of orbital pairs with mutual information above threshold that are far apart.
    """
    mi = get_all_mutual_information(mps)
    n_orbs = mi.shape[0]
    cost = 0.0
    for i in range(n_orbs):
        for j in range(i + 1, n_orbs):
            if mi[i, j] > threshold:
                cost += abs(i - j)
    return cost


def cost_crossing_mi_pairs(mps: MatrixProductState, threshold: float = 0.1) -> float:
    """
    Cost is the number of pairs of orbitals with mutual information above threshold that cross.
    Avoid high-MI pairs "crossing over" each other in the ordering
    """
    mi = get_all_mutual_information(mps)
    n_orbs = mi.shape[0]
    crossings = 0
    for i in range(n_orbs):
        for j in range(i + 1, n_orbs):
            if mi[i, j] < threshold:
                continue
            for k in range(i + 1, j):
                for l in range(j + 1, n_orbs):
                    if mi[k, l] > threshold and (i < k < j < l or k < i < l < j):
                        crossings += 1
    return crossings


def cost_entropy_max_to_mean(mps: MatrixProductState) -> float:
    """
    Cost based on the ratio of the maximum single-orbital entropy to the mean single-orbital entropy.
    Encourage a sharp entropy distribution
    → a few orbitals with high entanglement (to be kept in the active space)
    → and many with low entropy (to be discarded or treated classically)
    """
    entropies = [get_one_orbital_entropy(mps, i + 1) for i in range(mps.num_sites // 2)]
    mean = np.mean(entropies)
    return max(entropies) / mean if mean != 0 else -np.inf


def cost_function_to_dict(cost_function: Callable, **kwargs) -> dict[str, float]:
    function_params = kwargs
    match cost_function.__name__:
        case "cost_entropy":
            num_orbitals = function_params["num_orbitals"]
            return cost_entropy_dict(num_orbitals=num_orbitals)
        case "cost_total_mutual_information":
            num_orbitals = function_params["num_orbitals"]
            return cost_total_mutual_information_dict(num_orbitals=num_orbitals)
        case "cost_mutual_info_decay":
            num_orbitals = function_params["num_orbitals"]
            decay_power = function_params["decay_power"]
            return cost_mutual_info_decay_dict(num_orbitals=num_orbitals, decay_power=decay_power)
        case _:
            raise ValueError
    return


def calculate_purity(density_matrix: ndarray) -> float:
    rho_squared = density_matrix @ density_matrix
    purity = np.trace(rho_squared)
    return purity


def cost_function_dict_to_callable(
    cost_function_dict: dict[str, float], entropy_function: Callable[[ndarray], float]
) -> Callable[[MatrixProductState], float]:
    def cost_function(mps: MatrixProductState):
        cost = 0.0
        for s, weight in cost_function_dict.items():
            s_split = s.split("_")
            if s_split[0] == "S1":
                orbital_idx = int(s_split[1])
                rdm1 = get_one_orbital_rdm(mps, orbital_idx)
                cost += entropy_function(rdm1) * weight
            else:
                orbital_idx1 = int(s_split[1])
                orbital_idx2 = int(s_split[2])
                rdm2 = get_two_orbital_rdm(mps, [orbital_idx1, orbital_idx2])
                cost += entropy_function(rdm2) * weight
        return cost

    return cost_function


def cost_function_dict_to_purity_mpo(
    num_sites: int, cost_function_dict: dict[str, float]
) -> MatrixProductOperator:
    mpos = []
    for s, weight in cost_function_dict.items():
        s_split = s.split("_")
        if s_split[0] == "S1":
            orbital_idx = int(s_split[1])
            spin_orbitals = [2 * orbital_idx - 1, 2 * orbital_idx]
            temp_mpo = MatrixProductOperator.purity_mpo(num_sites, spin_orbitals)
            temp_mpo.multiply_by_constant(weight)
            mpos.append(temp_mpo)
        else:
            orbital_idx1 = int(s_split[1])
            orbital_idx2 = int(s_split[2])
            spin_orbitals = [
                2 * orbital_idx1 - 1,
                2 * orbital_idx1,
                2 * orbital_idx2 - 1,
                2 * orbital_idx2,
            ]
            temp_mpo = MatrixProductOperator.purity_mpo(num_sites, spin_orbitals)
            temp_mpo.multiply_by_constant(weight)
            mpos.append(temp_mpo)

    mpo = mpos[0]
    for next_mpo in mpos[1:]:
        mpo = mpo + next_mpo

    return mpo



import copy
import heapq
from timeit import default_timer

import numpy as np
from numpy import ndarray
from qiskit import QuantumCircuit
from scipy.sparse.linalg import LinearOperator, eigsh

from ...dmrg import DMRG
from ...mpo import MatrixProductOperator
from ...mps import MatrixProductState
from ...tn_methods.hf_suppression import HFSuppression
from ...tn_methods.mps_to_circuit import MPStoCircuit
from ..backend.base import QuantumBackend
from ..backend.tn_backend import TNQuantumBackend
from ..base import QuantumAlgorithm
from ..result import Result


class QSCI(QuantumAlgorithm):
    def __init__(
        self,
        hamiltonian: dict[str, complex],
        hf_state: str | None = None,
        backend: QuantumBackend | None = None,
        hfs: bool = True,
    ) -> "QSCI":
        """
        Constructor for QSCI class.
        """
        self.hamiltonian = hamiltonian
        self.hf_state = hf_state
        if hf_state is not None:
            self.num_electrons = hf_state.count("1")
            self.num_qubits = len(hf_state)
        else:
            self.num_electrons = None
            self.num_qubits = None
        self.hfs = hfs
        self.hamiltonian_mpo = MatrixProductOperator.from_hamiltonian(self.hamiltonian)
        self.state, self.energy = self.run_dmrg(self.hamiltonian)
        self._circuit = None
        self.set_backend(backend=backend)

    @property
    def circuit(self) -> QuantumCircuit:
        return self._circuit

    def run_dmrg(
        self, hamiltonian: dict, max_bond: int = 8, maxiter: int = 6
    ) -> MatrixProductState:
        """Run DMRG"""
        hf_mps = None
        if self.hf_state is not None:
            hf_mps = MatrixProductState.from_bitstring(self.hf_state)
            hf_mps = hf_mps.expand_bond_dimension_list(
                1, list(range(1, self.num_qubits))
            )
        dmrg = DMRG(hamiltonian, max_mps_bond=max_bond, initial_state=hf_mps)
        dmrg.run(maxiter=maxiter)
        return dmrg.mps, dmrg.energy

    def hf_suppression(
        self, mps: MatrixProductState, mpo: MatrixProductOperator | None = None
    ) -> tuple[MatrixProductState, MatrixProductOperator]:
        """Perform HF suppression"""
        hfs = HFSuppression(self.hf_state, mps, mpo)
        hfs.run()
        return hfs.suppressed_mps, hfs.evolved_mpo

    def prepare_state(self, mps: MatrixProductState) -> QuantumCircuit:
        """Prepare an MPS reference on quantum device"""
        circ = MPStoCircuit(mps, 1, 1.0).run()
        return circ

    def configuration_recovery(
        self, counts: dict[str, int], particle_number: int | None = None
    ) -> dict:
        """Perform configuration recovery"""
        new_counts = {k: v for k, v in counts.items() if v > 2}
        if particle_number is None:
            return new_counts
        else:
            new_counts = {
                k: v for k, v in new_counts.items() if k.count("1") == particle_number
            }
            return new_counts

    def gather_samples(self, cr_counts: dict, k: int) -> list[str]:
        """Collect the (at most) k most frequent samples to form the selected subspace"""
        top_samples = heapq.nlargest(k, cr_counts, key=cr_counts.get)
        return top_samples

    def project_hamiltonian(self, samples: list[str]) -> ndarray:
        """Project Hamiltonian onto subspace"""
        ham_mpo = copy.deepcopy(self.hamiltonian_mpo)
        basis = [MatrixProductState.from_bitstring(s) for s in samples]
        n = len(basis)
        ham_proj = np.zeros((n, n), dtype=complex)

        for i in range(n):
            h_i = basis[i].apply_mpo(ham_mpo)
            for j in range(i, n):
                h_ij = h_i.compute_inner_product(basis[j])
                ham_proj[i, j] = h_ij
                if i != j:
                    ham_proj[j, i] = h_ij.conjugate()

        return ham_proj

    def exact_diagonalisation(
        self, hamiltonian_matrix: ndarray
    ) -> tuple[float, ndarray]:
        """Perform exact diagonalisation on the projected Hamiltonian"""
        eval, evec = eigsh(hamiltonian_matrix, k=1, which="SA")
        return eval[0], evec[:, 0]

    def linear_operator_diagonalisation(
        self, samples: list[str]
    ) -> tuple[float, ndarray]:
        """For larger subspaces this will be more efficient"""
        basis_states = [MatrixProductState.from_bitstring(s) for s in samples]
        n = len(basis_states)
        ham_mpo = copy.deepcopy(self.hamiltonian_mpo)

        def matvec(v):
            psi_v = basis_states[0]
            psi_v.multiply_by_constant(v[0])
            for j in range(1, n):
                temp_mps = basis_states[j]
                temp_mps.multiply_by_constant(v[j])
                psi_v = psi_v + temp_mps
            H_psi_v = psi_v.apply_mpo(ham_mpo)
            return np.array(
                [basis_states[i].compute_inner_product(H_psi_v) for i in range(n)]
            )

        H_linear = LinearOperator(shape=(n, n), matvec=matvec, dtype=np.complex128)
        eval, evec = eigsh(H_linear, k=1, which="SA")
        return eval[0], evec[:, 0]

    def reconstruct_mps(
        self, samples: list[str], groundstate_vec: ndarray
    ) -> MatrixProductState:
        """Construct an MPS from the approximate groundstate solution"""
        new_mps = MatrixProductState.from_bitstring(samples[0])
        new_mps.multiply_by_constant(groundstate_vec[0])
        for i in range(1, len(samples)):
            temp_mps = MatrixProductState.from_bitstring(samples[i])
            temp_mps.multiply_by_constant(groundstate_vec[i])
            new_mps = new_mps + temp_mps

        return new_mps

    def run(
        self, num_shots: int, subspace_size: int, num_iterations: int = 1
    ) -> Result:
        """Run the full algorithm pipeline. Returns result object or final value."""
        start_time = default_timer()
        for _ in range(num_iterations):
            if self.hfs:
                self.state, _ = self.hf_suppression(self.state)
            self._circuit = self.prepare_state(self.state)
            counts = self.backend.run(self._circuit, shots=num_shots)
            cr_counts = self.configuration_recovery(counts, self.num_electrons)
            samples = self.gather_samples(cr_counts, subspace_size)
            if self.hfs and self.hf_state not in samples:
                samples += self.hf_state
            if len(samples) <= 500:
                projected_ham = self.project_hamiltonian(samples)
                self.energy, groundstate_vec = self.exact_diagonalisation(projected_ham)
            else:
                self.energy, groundstate_vec = self.linear_operator_diagonalisation(
                    samples
                )
            self.state = self.reconstruct_mps(samples, groundstate_vec)
        end_time = default_timer()

        metadata = {
            "algorithm_name": "QSCI",
            "num_shots": num_shots,
            "num_iterations": num_iterations,
            "max_subspace_size": subspace_size,
            "actual_subspace_size": len(samples),
            "total_runtime": end_time - start_time,
        }
        if self.backend is not None:
            metadata["backend_name"] = self.backend.name
            metadata["backend_coupling_map"] = self.backend.coupling_map
            metadata["backend_basis_gates"] = self.backend.basis_gates
            metadata["backend_num_qubits"] = self.backend.num_qubits

        result = self.energy

        result = Result(
            result=result,
            measurements=counts,
            parameters=None,
            metadata=metadata,
        )
        return result

    def set_backend(self, backend: QuantumBackend | None) -> None:
        """Attach a QuantumBackend instance for execution."""
        if backend is None:
            backend = TNQuantumBackend()
        self.backend = backend
        return



import copy
import inspect
from typing import Callable, List, TypeAlias, Union

import numpy as np
from numpy import ndarray
from qiskit import QuantumCircuit
from qiskit.circuit import CircuitInstruction, Operation
from qiskit.circuit.library import UnitaryGate
from sparse import SparseArray

from .backend.base import QuantumBackend

QiskitOptions: TypeAlias = Union[QuantumCircuit, Operation, CircuitInstruction]  # type: ignore
ArrayOptions = TypeAlias = Union[ndarray, SparseArray]


def count_qubits(obj: QiskitOptions | ArrayOptions) -> int:  # type: ignore
    """
    Count the number of qubits from an object.

    Args:
        obj: The circuit or array to count the qubits of

    Returns:
        The number of qubits acted on by obj
    """
    if isinstance(obj, QiskitOptions):
        num_qubits = obj.num_qubits
    elif isinstance(obj, ArrayOptions):
        num_qubits = int(np.log2(obj.shape[0]))

    return num_qubits


def to_quantum_circuit(obj: QiskitOptions | ArrayOptions) -> QuantumCircuit:  # type: ignore
    """
    Convert an object to a QuantumCircuit.

    Args:
        obj: The instruction or array to convert to a QuantumCircuit

    Returns:
        A QuantumCircuit representation of obj
    """
    num_qubits = count_qubits(obj)

    if isinstance(obj, QuantumCircuit):
        return obj
    elif isinstance(obj, Operation | CircuitInstruction):
        qc = QuantumCircuit(num_qubits)
        qc.append(obj, range(num_qubits))
        return qc
    elif isinstance(obj, ndarray):
        qc = QuantumCircuit(num_qubits)
        qc.append(UnitaryGate(obj), list(range(num_qubits))[::-1])
        return qc
    elif isinstance(obj, SparseArray):
        qc = QuantumCircuit(num_qubits)
        qc.append(UnitaryGate(obj.todense()), list(range(num_qubits))[::-1])
        return qc


def add_controls(qc: QuantumCircuit, ctrl_idxs: List[int]) -> QuantumCircuit:
    """
    Replace every instruction in qc with a controlled instruction on ctrl_idx.

    Args:
        qc: A quantum circuit
        ctrl_idxs: The indices of qubits to act as control bits

    Returns:
        A new QuantumCircuit
    """
    ctrl_qc = QuantumCircuit(qc.num_qubits)
    for inst in qc.data:
        qubits = [inst.qubits[i]._index for i in range(len(inst.qubits))][
            ::-1
        ] + ctrl_idxs
        ctrl_inst = inst.operation.control(len(ctrl_idxs))
        ctrl_qc.append(ctrl_inst, qubits[::-1])
    return ctrl_qc


def exp_pauli_string_to_circ(pauli_string: str, rot_angle: float) -> QuantumCircuit:
    """
    Create a circuit for an exponential Pauli string.

    Args:
        pauli_string: The Pauli string
        rot_angle: The rotation angle in the exponential

    Returns:
        A QuantumCircuit
    """
    qc = QuantumCircuit(len(pauli_string))

    if pauli_string == "I" * len(pauli_string):
        return qc

    for p_idx in range(len(pauli_string)):
        p = pauli_string[p_idx]
        if p == "X":
            qc.h(p_idx)
        elif p == "Y":
            qc.sdg(p_idx)
            qc.h(p_idx)

    non_id_qubits = [
        p_idx for p_idx in range(len(pauli_string)) if pauli_string[p_idx] != "I"
    ]
    for non_id_idx in range(len(non_id_qubits) - 1):
        q1, q2 = non_id_qubits[non_id_idx], non_id_qubits[non_id_idx + 1]
        qc.cx(q1, q2)
    qc.rz(2 * rot_angle, non_id_qubits[-1])
    for non_id_idx in range(len(non_id_qubits) - 1):
        q1, q2 = non_id_qubits[non_id_idx], non_id_qubits[non_id_idx + 1]
        qc.cx(q1, q2)

    for p_idx in range(len(pauli_string)):
        p = pauli_string[p_idx]
        if p == "X":
            qc.h(p_idx)
        elif p == "Y":
            qc.h(p_idx)
            qc.s(p_idx)

    return qc


def controlled_exp_pauli_string_circ(
    pauli_string: str, rot_angle: float, ctrl_idxs: list[int]
) -> QuantumCircuit:
    """
    Create a controlled circuit for an exponential Pauli string.

    Args:
        pauli_string: The Pauli string
        rot_angle: The rotation angle in the exponential
        ctrl_idxs: The list of qubits to act as controls

    Returns:
        A QuantumCircuit
    """
    qc = QuantumCircuit(len(pauli_string))

    for p_idx in range(len(pauli_string)):
        p = pauli_string[p_idx]
        if p == "X":
            qc.h(p_idx)
        elif p == "Y":
            qc.sdg(p_idx)
            qc.h(p_idx)

    non_id_qubits = [
        p_idx for p_idx in range(len(pauli_string)) if pauli_string[p_idx] != "I"
    ]
    for non_id_idx in range(len(non_id_qubits) - 1):
        q1, q2 = non_id_qubits[non_id_idx], non_id_qubits[non_id_idx + 1]
        qc.cx(q1, q2)
    qc.crz(2 * rot_angle, ctrl_idxs, non_id_idx[-1])
    for non_id_idx in range(len(non_id_qubits) - 1):
        q1, q2 = non_id_qubits[non_id_idx], non_id_qubits[non_id_idx + 1]
        qc.cx(q1, q2)

    for p_idx in range(len(pauli_string)):
        p = pauli_string[p_idx]
        if p == "X":
            qc.h(p_idx)
        elif p == "Y":
            qc.h(p_idx)
            qc.s(p_idx)

    return qc


def calculate_exp_val(
    circuit: QuantumCircuit, observable: dict, backend: QuantumBackend, shots: int
) -> float:
    """Calculate an expectation value from a circuit

    Args:
        circuit: The QuantumCircuit
        observable: The observable (sum of Pauli terms)
        backend: The backend
        shots: Shots per circuit
    """

    def filter_counts(counts: dict, relevant_qidxs: list[int]):
        new_counts = {}
        for k, v in counts.items():
            new_k = ""
            for b_idx in range(len(k)):
                b = k[b_idx]
                if b_idx in relevant_qidxs:
                    new_k += b
            if new_k in new_counts:
                new_counts[new_k] += v
            else:
                new_counts[new_k] = v
        return new_counts

    def parity_of_bitstring(bitstring: str):
        parity = 1
        for b in bitstring:
            if b == "1":
                parity *= -1
        return parity

    exp_val = 0
    for k, v in observable.items():
        if k == "I" * len(k):
            exp_val += v
            continue

        qc = copy.deepcopy(circuit)
        relevant_qidxs = []
        for p_idx in range(len(k)):
            p = k[p_idx]
            if p == "X":
                qc.h(p_idx)
                relevant_qidxs.append(p_idx)
            elif p == "Y":
                qc.sdg(p_idx)
                qc.h(p_idx)
                relevant_qidxs.append(p_idx)
            elif p == "Z":
                relevant_qidxs.append(p_idx)
        counts = backend.run(qc, shots=shots)
        filtered_counts = filter_counts(counts, relevant_qidxs)
        counts_p = 0
        counts_n = 0
        for a, b in filtered_counts.items():
            parity = parity_of_bitstring(a)
            if parity == 1:
                counts_p += b
            else:
                counts_n += b
        exp_val += v * ((counts_p - counts_n) / shots)

    return exp_val


def calculate_exp_val_stochastic(
    circuit_builder: Callable,
    observable: dict,
    backend: QuantumBackend,
    shots: int,
    *args,
    **kwargs,
) -> float:
    """Calculate an expectation value from a circuit

    Args:
        circuit_builder: The QuantumCircuit builder function
        observable: The observable (sum of Pauli terms)
        backend: The backend
        shots: Shots per circuit
        args: For circuit builder function
        kwargs: For circuit builder function
    """

    def filter_counts(counts: dict, relevant_qidxs: list[int]):
        new_counts = {}
        for k, v in counts.items():
            new_k = ""
            for b_idx in range(len(k)):
                b = k[b_idx]
                if b_idx in relevant_qidxs:
                    new_k += b
            if new_k in new_counts:
                new_counts[new_k] += v
            else:
                new_counts[new_k] = v
        return new_counts

    def parity_of_bitstring(bitstring: str):
        parity = 1
        for b in bitstring:
            if b == "1":
                parity *= -1
        return parity

    exp_val = 0
    for k, v in observable.items():
        if k == "I" * len(k):
            exp_val += v
            continue

        counts = {}
        for _ in range(shots):
            sig = inspect.signature(circuit_builder)
            if len(sig.parameters) == 0:
                qc = circuit_builder()
            else:
                qc = circuit_builder(*args, **kwargs)
            relevant_qidxs = []
            for p_idx in range(len(k)):
                p = k[p_idx]
                if p == "X":
                    qc.h(p_idx)
                    relevant_qidxs.append(p_idx)
                elif p == "Y":
                    qc.sdg(p_idx)
                    qc.h(p_idx)
                    relevant_qidxs.append(p_idx)
                elif p == "Z":
                    relevant_qidxs.append(p_idx)
            bs = list(backend.run(qc, shots=1).keys())[0]
            if bs in counts:
                counts[bs] += 1
            else:
                counts[bs] = 1

        filtered_counts = filter_counts(counts, relevant_qidxs)
        counts_p = 0
        counts_n = 0
        for a, b in filtered_counts.items():
            parity = parity_of_bitstring(a)
            if parity == 1:
                counts_p += b
            else:
                counts_n += b
        exp_val += v * ((counts_p - counts_n) / shots)

    return exp_val



import numpy as np
from numpy import ndarray

from .mpo import MatrixProductOperator
from .mps import MatrixProductState

# NOTATION:
# The following is a matrix of one orbital operators.
# Each matrix element is a linear combination of strings of creation and annihilation operators given as a list of tuples.
# Each tuple is (list, weight) where the first term in list is the operators acting on the up spin-orbital and the second term is the operators acting on the down spin-orbital
ONE_ORBITAL_OPERATORS = np.array(
    [
        [
            [(["", ""], 1), (["+-", ""], -1), (["", "+-"], -1), (["+-", "+-"], 1)],
            [(["", "-"], 1), (["+-", "-"], -1)],
            [(["-", ""], 1), (["-", "+-"], -1)],
            [(["-", "-"], 1)],
        ],
        [
            [(["", "+"], 1), (["+-", "+"], -1)],
            [(["", "+-"], 1), (["+-", "+-"], -1)],
            [(["-", "+"], 1)],
            [(["-", "+-"], -1)],
        ],
        [
            [(["+", ""], 1), (["+", "+-"], -1)],
            [(["+", "-"], 1)],
            [(["+-", ""], 1), (["+-", "+-"], -1)],
            [(["+-", "-"], 1)],
        ],
        [
            [(["+", "+"], 1)],
            [(["+", "+-"], 1)],
            [(["+-", "+"], 1)],
            [(["+-", "+-"], 1)],
        ],
    ],
    dtype=object,
).reshape((4, 4))


def two_orbital_basis_reorder(idx: int, particle_to_orbital: bool = True) -> int:
    """
    The basis for a two-orbital RDM can be ordered particle-first or orbital-first.
    Particle-first makes the two-orbital RDM block diagonal however we tend to use orbital-first.

    Args:
        idx: The index of the basis element.
        particle_to_orbital: If true, finds the index of the basis element in the orbital basis.

    Returns:
        int: The index of the basis element in the reordered basis.
    """
    # key = index in particle-first ordering, value = index in orbital-first ordering
    reordering_dict = {
        "0": "0",
        "1": "2",
        "2": "8",
        "3": "1",
        "4": "4",
        "5": "10",
        "6": "3",
        "7": "9",
        "8": "6",
        "9": "12",
        "10": "5",
        "11": "11",
        "12": "14",
        "13": "7",
        "14": "13",
        "15": "15",
    }
    # If reordering in the other direction, flip the dictionary
    if not particle_to_orbital:
        reordering_dict = {v: k for k, v in reordering_dict.items()}
    return int(reordering_dict[str(idx)])


def get_one_orbital_operator(idx1: int, idx2: int, orbital_idx: int) -> list[tuple]:
    """
    Get a one orbital Fermionic operator.

    Args:
        idx1: The index of the iniital occupation state.
        idx2: The index of the final occupation state.
        orbital_idx: The index of the orbital.

    Returns:
        A list of tuples of the form (op, weight) where op is a single string of Fermionic creation/annihilation operators.
    """
    up_idx = 2 * orbital_idx - 2
    down_idx = 2 * orbital_idx - 1
    ops = []
    for op_list, weight in ONE_ORBITAL_OPERATORS[idx1, idx2]:
        op = []
        up_op = op_list[0]
        down_op = op_list[1]
        for o in up_op:
            op.append((str(up_idx), o))
        for o in down_op:
            op.append((str(down_idx), o))
        ops.append((op, weight))
    return ops


def get_one_particle_rdm(mps: MatrixProductState) -> ndarray:
    """Calculate 1-RDM.

    Args:
        mps: The quantum state whose 1-RDM we want

    Returns:
        An array of shape (N,N)
    """
    rdm = np.zeros((mps.num_sites, mps.num_sites))

    # We need to calculate pairs (i,j) for i <= j
    for j in range(mps.num_sites):
        for i in range(j + 1):
            op_list = [(f"{i}", "+"), (f"{j}", "-")]
            mpo = MatrixProductOperator.from_fermionic_string(op_list)
            expval = mps.compute_expectation_value(mpo)
            rdm[i, j] = expval
            if i != j:
                rdm[j, i] = np.conj(expval)

    return rdm


def get_two_particle_rdm(mps: MatrixProductState) -> ndarray:
    """Calculate 2-RDM.

    Args:
        mps: The quantum state whose 2-RDM we want

    Returns:
        An array of shape (N,N,N,N)
    """
    rdm = np.zeros((mps.num_sites, mps.num_sites, mps.num_sites, mps.num_sites))

    # For pairs (pq),(rs) the 2-RDM is antisymmetric within each pair and Hermitian between the pairs.
    pairs = [(p, q) for q in range(mps.num_sites) for p in range(q + 1)]

    for pair1_idx in range(len(pairs)):
        for pair2_idx in range(pair1_idx):
            p, q = pairs[pair1_idx]
            r, s = pairs[pair2_idx]
            op_list = [
                (f"{p}", "+"),
                (f"{q}", "+"),
                (f"{r}", "-"),
                (f"{s}", "-"),
            ]
            mpo = MatrixProductOperator.from_fermionic_operator(op_list)
            expval = mps.compute_expectation_value(mpo)
            rdm[p, q, r, s] = expval
            rdm[q, p, r, s] = -expval
            rdm[p, q, s, r] = -expval
            rdm[r, s, p, q] = np.conj(expval)
            rdm[s, r, p, q] = -np.conj(expval)
            rdm[r, s, q, p] = -np.conj(expval)

    return rdm


def get_one_orbital_rdm(
    mps: MatrixProductState,
    orbital_idx: int,
    direct: bool = True,
    enforce_symmetry: bool = False,
) -> ndarray:
    """
    Calculate the one orbital RDM.

    Args:
        mps: The quantum state.
        site: The location of the orbital
        direct: If True the RDM is calculated via direct contraction, otherwise calculated via matrix elements.
        enforce_symmetry: If True then enforces spin and particle number symmetry in RDM.

    Return:
        The (4,4) array for the one-orbital RDM.

    Example:
        >>> mps = MatrixProductState(...)
        >>> rdm = get_one_orbital_rdm(mps, 1)
        >>> print(rdm)
    """
    if direct:
        spin_orbitals_to_remove = list(range(1, mps.num_sites + 1))
        spin_orbitals_to_remove.remove(2 * orbital_idx - 1)
        spin_orbitals_to_remove.remove(2 * orbital_idx)
        rdm = mps.partial_trace(spin_orbitals_to_remove, matrix=True)
        rdm = rdm.data.todense()
        if enforce_symmetry:
            for j in range(4):
                for i in range(j):
                    rdm[i, j] = 0
                    rdm[j, i] = 0
        return rdm
    else:
        rdm = np.zeros((4, 4), dtype=complex)
        for i in range(4):
            op = get_one_orbital_operator(i, i, orbital_idx)
            mpo = MatrixProductOperator.from_fermionic_operator(mps.num_sites, op)
            rdm[i, i] = mps.compute_expectation_value(mpo)
        if enforce_symmetry:
            return rdm
        else:
            for i in range(4):
                for j in range(4):
                    if i == j:
                        pass
                    else:
                        op = get_one_orbital_operator(i, j, orbital_idx)
                        mpo = MatrixProductOperator.from_fermionic_operator(
                            mps.num_sites, op
                        )
                        rdm[i, j] = mps.compute_expectation_value(mpo)
            return rdm


def get_two_orbital_rdm(
    mps: MatrixProductState,
    sites: list[int],
    direct: bool = True,
    enforce_symmetry: bool = False,
) -> ndarray:
    """Calculate the two orbital RDM.

    Args:
        mps: The quantum state.
        site: The location of the orbitals
        direct: If True the RDM is calculated via direct contraction, otherwise calculated via matrix elements.
        enforce_symmetry: If True then enforces spin and particle number symmetry in RDM.

    Return:
        The (16,16) array for the two-orbital RDM

    Raises:
        ValueError: If the sites are not two different sites.

    Example:
        >>> mps = MatrixProductState(...)
        >>> rdm = get_two_orbital_rdm(mps, [1, 2])
        >>> print(rdm)
    """
    if len(set(sites)) != 2:
        raise ValueError("Must provide two different sites.")

    expected_non_zero_pairs_particle_basis = [
        (0, 0),
        (1, 1),
        (1, 2),
        (2, 1),
        (2, 2),
        (3, 3),
        (3, 4),
        (4, 3),
        (4, 4),
        (5, 5),
        (6, 6),
        (6, 7),
        (6, 8),
        (6, 9),
        (7, 6),
        (7, 7),
        (7, 8),
        (7, 9),
        (8, 6),
        (8, 7),
        (8, 8),
        (8, 9),
        (9, 6),
        (9, 7),
        (9, 8),
        (9, 9),
        (10, 10),
        (11, 11),
        (11, 12),
        (12, 11),
        (12, 12),
        (13, 13),
        (13, 14),
        (14, 13),
        (14, 14),
        (15, 15),
    ]
    expected_non_zero_pairs_orbital_basis = [
        (two_orbital_basis_reorder(x), two_orbital_basis_reorder(y))
        for x, y in expected_non_zero_pairs_particle_basis
    ]
    if direct:
        spin_orbitals_to_remove = list(range(1, mps.num_sites + 1))
        spin_orbitals_to_remove.remove(2 * sites[0] - 1)
        spin_orbitals_to_remove.remove(2 * sites[0])
        spin_orbitals_to_remove.remove(2 * sites[1] - 1)
        spin_orbitals_to_remove.remove(2 * sites[1])
        rdm = mps.partial_trace(spin_orbitals_to_remove, matrix=True)
        rdm = rdm.data.todense()
        if enforce_symmetry:
            for i in range(16):
                for j in range(16):
                    if (i, j) in expected_non_zero_pairs_orbital_basis:
                        continue
                    else:
                        rdm[i, j] = 0
        return rdm
    else:
        rdm = np.zeros((16, 16), dtype=complex)
        for i in range(16):
            for j in range(16):
                if (
                    i,
                    j,
                ) not in expected_non_zero_pairs_orbital_basis and enforce_symmetry:
                    rdm[i, j] = 0
                else:
                    op1 = get_one_orbital_operator(
                        int(np.floor(i / 4)), int(np.floor(j / 4)), sites[0]
                    )
                    op2 = get_one_orbital_operator(int(i % 4), int(j % 4), sites[1])
                    mpo1 = MatrixProductOperator.from_fermionic_operator(
                        mps.num_sites, op1
                    )
                    mpo2 = MatrixProductOperator.from_fermionic_operator(
                        mps.num_sites, op2
                    )
                    mpo = mpo1 * mpo2
                    rdm[i, j] = mps.compute_expectation_value(mpo)
        return rdm


def get_one_orbital_entropy(mps: MatrixProductState, site: int) -> float:
    """Calculate the one orbital entropy.

    Args:
        mps (MatrixProductState): A matrix product state.
        site (int): The index of the orbital (1-indexed).
    Returns:
        float: The one orbital entropy.

    Example:
        >>> mps = MatrixProductState(...)
        >>> entropy = get_one_orbital_entropy(mps, 1)
        >>> print(entropy)
    """
    rdm1 = get_one_orbital_rdm(mps, site, direct=True, enforce_symmetry=False)
    # Calculate eigenvalues
    eigvals = np.linalg.eigvalsh(rdm1)
    # Calculate entropy
    entropy = -np.sum(
        eigvals * np.log2(eigvals + 1e-12)
    )  # Add small value to avoid log(0)
    return entropy


def get_two_orbital_entropy(mps: MatrixProductState, sites: list[int]) -> float:
    """Calculate the two orbital entropy.

    Args:
        mps (MatrixProductState): A matrix product state.
        sites (list[int]): A list of two orbital indices (1-indexed).
    Returns:
        float: The two orbital entropy.

    Example:
        >>> mps = MatrixProductState(...)
        >>> entropy = get_two_orbital_entropy(mps, [1, 2])
        >>> print(entropy)
    """
    match n_sites := len(set(sites)):
        case 1:
            return get_one_orbital_entropy(mps, sites[0])
        case 2:
            rdm2 = get_two_orbital_rdm(mps, sites, direct=True, enforce_symmetry=False)
            # Calculate eigenvalues
            eigvals = np.linalg.eigvalsh(rdm2)
            # Calculate entropy
            entropy = -np.sum(
                eigvals * np.log2(eigvals + 1e-12)
            )  # Add small value to avoid log(0)
            return entropy
        case _:
            raise ValueError(
                "Incorrect number of sites provided for two orbital entropy %s", n_sites
            )


def get_mutual_information(mps: MatrixProductState, sites: list[int]) -> float:
    """
    Calculate the mutual information between two orbitals.
    I(i, j) = S(i) + S(j) - S(i,j)

    Args:
        mps (MatrixProductState): A matrix product state.
        sites (list[int]): A list of two orbital indices (1-indexed).
    Returns:
        float: The mutual information between the two orbitals.

    Raises:
        ValueError: If the sites list does not contain exactly two different sites.

    Example:
        >>> mps = MatrixProductState(...)
        >>> mutual_info = get_mutual_information(mps, [1, 2])
        >>> print(mutual_info)
    """
    match n_sites := len(set(sites)):
        case 1:
            mutual_info = get_one_orbital_entropy(mps, sites[0])
        case 2:
            s1 = get_one_orbital_entropy(mps, sites[0])
            s2 = get_one_orbital_entropy(mps, sites[1])
            s12 = get_two_orbital_entropy(mps, sites)
            mutual_info = s1 + s2 - s12
        case _:
            raise ValueError(
                "Incorrect number of sites provided for mutual information %s", n_sites
            )
    return mutual_info


def get_all_mutual_information(mps: MatrixProductState) -> np.typing.NDArray:
    """
    Calculate the mutual information between every pair of orbitals.
    Mutual Information matrix where M[i,j] = I(i,j)
    I(i,j) = S(i) + S(j) - S(i,j)

    Args:
        mps (MatrixProductState): A matrix product state.
    Returns:
        ndarray: A symmetric matrix of shape (N, N) where N is the number of orbitals.
                 M[i, j] = I(i, j) is the mutual information between orbitals i and j.

    Example:
        >>> mps = MatrixProductState(...)
        >>> mutual_info_matrix = get_all_mutual_information(mps)
        >>> print(mutual_info_matrix)
    """
    n_orbs = mps.num_sites // 2  # Number of orbitals
    M = np.zeros((n_orbs, n_orbs))
    for i in range(0, n_orbs):
        M[i, i] = get_one_orbital_entropy(mps, i + 1)
        for j in range(i + 1, n_orbs):
            # mps sites are 1-indexed
            M[i, j] = get_mutual_information(mps, [i + 1, j + 1])
            M[j, i] = M[i, j]
    return M



import numpy as np
from qiskit import QuantumCircuit
from qiskit.circuit import Parameter
from qiskit.circuit.library import (
    CXGate,
    CZGate,
    ExcitationPreserving,
    PauliTwoDesign,
    RXGate,
    RYGate,
    RZGate,
)
from qiskit.quantum_info import Operator, random_unitary


def hea_ansatz(
    n_qubits: int, layers: int, sq_rotations: list[str], mq_gate: str
) -> QuantumCircuit:
    """
    Hardware Efficient Ansatz.

    Args:
        n_qubits: The number of qubits
        layers: The number of layers in the ansatz
        sq_rotations: The list of single-qubit operations available on the hardware
        mq_gate: The multi-qubit (generally 2q) gate available on the hardware

    Returns:
        QuantumCircuit
    """
    hea = QuantumCircuit(n_qubits)
    num_params = 3 * n_qubits * layers
    params = [Parameter(str(a)) for a in range(num_params)]

    def get_sq_gate(idx, param):
        sq = sq_rotations[idx]
        if "rx" == sq:
            return RXGate(param)
        if "ry" == sq:
            return RYGate(param)
        if "rz" == sq:
            return RZGate(param)

    if mq_gate == "cx":
        mq_gate = CXGate()
    elif mq_gate == "cz":
        mq_gate = CZGate()

    for _ in range(layers):
        for q_idx in range(n_qubits):
            gate = get_sq_gate(0, params[0])
            hea.append(gate, q_idx)
            params.pop(0)
        for q_idx in range(n_qubits):
            gate = get_sq_gate(1, params[0])
            hea.append(gate, q_idx)
            params.pop(0)
        for q_idx in range(n_qubits):
            gate = get_sq_gate(0, params[0])
            hea.append(gate, q_idx)
            params.pop(0)

        for q_idx in range(n_qubits - 1):
            hea.append(mq_gate, (q_idx, q_idx + 1))
        hea.append(mq_gate, (n_qubits - 1, 0))

    return hea


def pauli_two_design_ansatz(n_qubits: int, layers: int) -> QuantumCircuit:
    """
    Pauli two design ansatz.

    Args:
        n_qubits: The number of qubits
        layers: The number of layers in the ansatz

    Returns:
        QuantumCircuit
    """
    return PauliTwoDesign(n_qubits, reps=layers).decompose()


def number_preserving_ansatz(
    n_qubits: int, layers: int, entanglement: str
) -> QuantumCircuit:
    """
    Number preserving ansatz.

    Args:
        n_qubits: The number of qubits
        layers: The number of layers in the ansatz
        entanglement: What entanglement structure to use

    Returns:
        QuantumCircuit
    """
    return ExcitationPreserving(
        n_qubits, reps=layers, entanglement=entanglement
    ).decompose()


def identity_brickwork_circuit_offset(n_qubits: int, layers: int) -> QuantumCircuit:
    """
    Brickwork circuit coupling alpha-alpha electrons and beta-beta electrons

    Args:
        n_qubits: number of qubits
        layers: number of layers

    Returns:
        QuantumCircuit
    """
    id_gate = Operator(np.eye(4)).to_instruction()
    qc = QuantumCircuit(n_qubits)
    for _ in range(layers):
        for idx in range(int(n_qubits / 2)):
            qidxs = [2 * idx + x for x in [0, 2]]
            qc.append(id_gate, qidxs)
        for idx in range(int(n_qubits / 2)):
            qidxs = [2 * idx + x for x in [1, 3]]
            qc.append(id_gate, qidxs)
    return qc


def random_brickwork_circuit_offset(n_qubits: int, layers: int) -> QuantumCircuit:
    """
    Brickwork circuit of random gates coupling alpha-alpha electrons and beta-beta electrons

    Args:
        n_qubits: The number of qubits
        layers: The number of layers

    Returns:
        A quantum circuit
    """
    qc = QuantumCircuit(n_qubits)
    for _ in range(layers):
        for idx in range(int(n_qubits / 4)):
            random_gate = random_unitary(4)
            qidxs = [2 * idx + x for x in [0, 2]]
            qc.append(random_gate, qidxs)
        for idx in range(int(n_qubits / 4)):
            random_gate = random_unitary(4)
            qidxs = [2 * idx + x for x in [1, 3]]
            qc.append(random_gate, qidxs)
    return qc


def identity_brickwork_circuit(
    n_qubits: int, layers: int, qubits_per_gate: int = 2, gap: int = 1
) -> QuantumCircuit:
    """
    Brickwork circuit of identity gates.

    Args:
        n_qubits: The number of qubits
        layers: The number of layers
        qubits_per_gate: The number of qubits per gate
        gap: The number of qubits to skip between layers

    Returns:
        A quantum circuit
    """
    id_gate = Operator(np.eye(2**qubits_per_gate)).to_instruction()
    qc = QuantumCircuit(n_qubits)
    for _ in range(layers):
        for idx in range(int(np.floor(n_qubits / qubits_per_gate))):
            qidxs = [qubits_per_gate * idx + x for x in range(qubits_per_gate)]
            qc.append(id_gate, qidxs)
        for idx in range(int(np.floor((n_qubits - gap) / qubits_per_gate))):
            qidxs = [qubits_per_gate * idx + gap + x for x in range(qubits_per_gate)]
            qc.append(id_gate, qidxs)
    return qc


def random_brickwork_circuit(
    n_qubits: int, layers: int, qubits_per_gate: int = 2, gap: int = 1
) -> QuantumCircuit:
    """
    Brickwork circuit of random gates.

    Args:
        n_qubits: The number of qubits
        layers: The number of layers
        qubits_per_gate: The number of qubits per gate
        gap: The number of qubits to skip between layers

    Returns:
        A quantum circuit
    """
    qc = QuantumCircuit(n_qubits)
    for _ in range(layers):
        for idx in range(int(np.floor(n_qubits / qubits_per_gate))):
            random_gate = random_unitary(2**qubits_per_gate)
            qidxs = [qubits_per_gate * idx + x for x in range(qubits_per_gate)]
            qc.append(random_gate, qidxs)
        for idx in range(int(np.floor((n_qubits - gap) / qubits_per_gate))):
            random_gate = random_unitary(2**qubits_per_gate)
            qidxs = [qubits_per_gate * idx + gap + x for x in range(qubits_per_gate)]
            qc.append(random_gate, qidxs)
    return qc


def identity_staircase_circuit(
    n_qubits: int, layers: int, qubits_per_gate: int
) -> QuantumCircuit:
    """
    Staircase circuit of identity gates.

    Args:
        n_qubits: The number of qubits
        layers: The number of layers
        qubits_per_gate: The number of qubits per gate

    Returns:
        A quantum circuit
    """
    id_gate = Operator(np.eye(2**qubits_per_gate)).to_instruction()
    qc = QuantumCircuit(n_qubits)
    for _ in range(layers):
        for idx in range(n_qubits + 1 - qubits_per_gate):
            qidxs = [idx + x for x in range(qubits_per_gate)]
            qc.append(id_gate, qidxs)
    return qc


def random_staircase_circuit(
    n_qubits: int, layers: int, qubits_per_gate: int
) -> QuantumCircuit:
    """
    Staircase circuit of random gates.

    Args:
        n_qubits: The number of qubits
        layers: The number of layers
        qubits_per_gate: The number of qubits per gate

    Returns:
        A quantum circuit
    """
    qc = QuantumCircuit(n_qubits)
    for _ in range(layers):
        for idx in range(n_qubits + 1 - qubits_per_gate):
            random_gate = random_unitary(2**qubits_per_gate)
            qidxs = [idx + x for x in range(qubits_per_gate)]
            qc.append(random_gate, qidxs)
    return qc
